import importlib
from typing import List

from loguru import logger
import sys

from indox.core import Document
from indox.data_loaders.utils import convert_latex_to_md
from indox.vector_stores.utils import filter_complex_metadata

# Set up logging
logger.remove()  # Remove the default logger
logger.add(sys.stdout,
           format="<green>{level}</green>: <level>{message}</level>",
           level="INFO")

logger.add(sys.stdout,
           format="<red>{level}</red>: <level>{message}</level>",
           level="ERROR")


def import_unstructured_partition(content_type):
    # Import appropriate partition function from the `unstructured` library
    module_name = f"unstructured.partition.{content_type}"
    module = importlib.import_module(module_name)
    partition_function_name = f"partition_{content_type}"
    prt = getattr(module, partition_function_name)
    return prt


def create_documents_unstructured(file_path):
    try:
        if file_path.lower().endswith(".pdf"):
            # Partition PDF with a high-resolution strategy
            from unstructured.partition.pdf import partition_pdf
            elements = partition_pdf(
                filename=file_path,
                strategy="hi_res",
                # infer_table_structure=True,
            )
            # Remove "References" and header elements
            reference_title = [
                el for el in elements
                if el.text == "References" and el.category == "Title"
            ][0]
            references_id = reference_title.id
            elements = [el for el in elements if el.metadata.parent_id != references_id]
            elements = [el for el in elements if el.category != "Header"]
        elif file_path.lower().endswith(".xlsx"):
            from unstructured.partition.xlsx import partition_xlsx
            elements_ = partition_xlsx(filename=file_path)
            elements = [el for el in elements_ if el.metadata.text_as_html is not None]
        elif file_path.lower().startswith("www") or file_path.lower().startswith("http"):
            from unstructured.partition.html import partition_html
            elements = partition_html(url=file_path)
        else:
            if file_path.lower().endswith(".tex"):
                file_path = convert_latex_to_md(latex_path=file_path)
            content_type = file_path.lower().split(".")[-1]
            if content_type == "txt":
                prt = import_unstructured_partition(content_type="text")
            else:
                prt = import_unstructured_partition(content_type=content_type)
            elements = prt(filename=file_path)
        return elements
    except AttributeError as ae:
        logger.error(f"Attribute error: {ae}")
        return ae
    except Exception as e:
        logger.error(f"Unexpected error: {e}")
        return e


def get_chunks_unstructured(file_path, chunk_size, remove_sword, splitter):
    """
    Extract chunks from an unstructured document file using an unstructured data processing library.

    Parameters:
    - file_path (str): The path to the file containing unstructured data.
    - chunk_size (int, optional): The maximum size (in characters) for each chunk. Defaults to 500.

    Returns:
    - list: A list of `Document` objects, each containing a portion of the original content with relevant metadata.

    Raises:
    - Exception: Any errors that occur during document processing or chunking.

    Notes:
    - The function uses a title-based chunking method to segment the unstructured data into logical parts.
    - Metadata is cleaned and filtered to ensure proper structure before being added to the `Document` objects.
    - The `filter_complex_metadata` function is used to simplify and sanitize metadata attributes.

    """
    from unstructured.chunking.title import chunk_by_title

    try:
        logger.info("Starting processing")

        # Create initial document elements using the unstructured library
        elements = create_documents_unstructured(file_path)

        if splitter:
            logger.info("Using custom splitter")
            text = ""
            for el in elements:
                text += el.text

            documents = splitter(text=text, max_tokens=chunk_size)
        else:
            logger.info("Using title-based chunking")
            # Split elements based on the title and the specified max characters per chunk
            elements = chunk_by_title(elements, max_characters=chunk_size)
            documents = []

            # Convert each element into a `Document` object with relevant metadata
            for element in elements:
                metadata = element.metadata.to_dict()
                del metadata["languages"]  # Remove unnecessary metadata field

                for key, value in metadata.items():
                    if isinstance(value, list):
                        value = str(value)
                    metadata[key] = value

                if remove_sword:
                    from indox.data_loader_splitter.utils.clean import remove_stopwords
                    element.text = remove_stopwords(element.text)

                # documents.append(Document(page_content=element.text, metadata=**metadata))
                documents.append(Document(page_content=element.text.replace("\n", ""), **metadata))

            # Filter and sanitize complex metadata
            documents = filter_complex_metadata(documents=documents)

        logger.info("Completed chunking process")
        return documents

    except Exception as e:
        logger.error(f"Failed at step with error: {e}")
        raise


class Unstructured:
    def __init__(self, file_path: str):
        """
        Initialize the UnstructuredLoadAndSplit class.

        Parameters:
        - file_path (str): The path to the file containing unstructured data.

        """
        try:
            self.file_path = file_path
        except Exception as e:
            logger.error(f"Error initializing Unstructured: {e}")
            raise

    def load(self):

        elements = create_documents_unstructured(file_path=self.file_path)
        return elements

    def load_and_split(self, remove_stopwords: bool = False, max_chunk_size: int = 500, splitter=None) -> (
            List)['Document']:
        """
        Split an unstructured document into chunks.

        Returns:
        - List[Document]: A list of `Document` objects, each containing a portion of the original content with relevant metadata.
        """
        try:
            logger.info("Getting all documents")
            docs = get_chunks_unstructured(self.file_path, max_chunk_size, remove_stopwords, splitter)
            logger.info("Successfully obtained all documents")
            return docs
        except Exception as e:
            logger.error(f"Error in get_all_docs: {e}")
            raise



from .semantic_text_splitter import SemanticTextSplitter
from .recursively_split_charachter import RecursiveCharacterTextSplitter
from .split_charachter import CharacterTextSplitter
from .markdown_text_splitter import MarkdownTextSplitter
from .ai21_text_splitter import AI21SemanticTextSplitter



from typing import List, Dict, Tuple
from indox.core import Document
import os

class Deeplake:
    """
    A class to interface with a DeepLake-based vector store for storing and searching text data using embeddings.

    Attributes:
    - collection_name (str): The name of the collection to store in the vector store.
    - embedding_function (callable): A function that takes text input and returns an embedding vector.

    Methods:
    - __init__(collection_name, embedding_function):
        Initializes the Deeplake instance with a specified collection name and embedding function.

    - add(docs, metadata=None):
        Adds a list of documents and optional metadata to the vector store.

        Parameters:
        - docs (List[str]): A list of text documents to be added to the vector store.
        - metadata (List[Dict], optional): A list of dictionaries containing metadata for each document.
          If not provided, default metadata with the source file path is generated.

        Raises:
        - ValueError: If `docs` is not a list of strings or if `metadata` is not a list of dictionaries.

    - similarity_search_with_score(query, k=5):
        Performs a similarity search on the vector store with the provided query and returns the most similar documents along with their scores.

        Parameters:
        - query (str): The query text for which similar documents are to be found.
        - k (int, optional): The number of top results to return. Default is 5.

        Returns:
        - List[Tuple[Document, float]]: A list of tuples where each tuple contains a `Document` object and its corresponding similarity score.

    Raises:
    - FileNotFoundError: If the specified collection path does not exist.
    - RuntimeError: If there is an issue with the vector store operation.
    """
    def __init__(self, collection_name: str, embedding_function):
        self.collection_name = collection_name
        self.embedding_function = embedding_function
        from deeplake.core.vectorstore import VectorStore

        # Initialize the VectorStore with the specified collection name
        self.vector_store = VectorStore(
            path=f"/content/vector_store/{self.collection_name}"
        )

    def add(self, docs: List[str], metadata: List[Dict] = None):
        if not isinstance(docs, list) or not all(isinstance(text, str) for text in docs):
            raise ValueError("texts must be a list of strings")

        if metadata is None:
            # Generate default metadata if none is provided
            metadata = [{"source": os.path.abspath(doc)} for doc in docs]
        elif not isinstance(metadata, list) or not all(isinstance(meta, dict) for meta in metadata):
            raise ValueError("metadata must be a list of dictionaries")

        documents = [Document(page_content=doc, metadata=meta) for doc, meta in zip(docs, metadata)]

        self.vector_store.add(
            text=docs,
            embedding_function=self.embedding_function,
            embedding_data=docs,
            metadata=metadata
        )

    def similarity_search_with_score(self, query: str, k: int = 5) -> List[Tuple[Document, float]]:
        # Perform the search
        results = self.vector_store.search(
            embedding_data=query,
            embedding_function=self.embedding_function
        )

        scores = results['score']
        ids = results['id']
        metadata = results['metadata']
        texts = results['text']

        documents = [Document(page_content=text, metadata=meta) for text, meta in zip(texts, metadata)]

        return [(doc, score) for doc, score in zip(documents, scores)]




from indox.core import Document
from loguru import logger
import sys
import numpy as np
from typing import List, Tuple, Union
import json
import uuid

logger.remove()
logger.add(sys.stdout,
           format="<green>{level}</green>: <level>{message}</level>",
           level="INFO")
logger.add(sys.stdout,
           format="<red>{level}</red>: <level>{message}</level>",
           level="ERROR")


class RedisDB:
    """
    A concrete implementation of VectorStore using Redis for storage.

    Attributes:
        redis_client (redis.Redis): The Redis client.
        embedding (callable): The embedding function to use.
        prefix (str): The prefix for Redis keys.
    """

    def __init__(self, host: str, port: int, password: str, embedding: callable, prefix: str = "doc"):
        """
        Initializes the RedisVector.

        Args:
            host (str): The host of the Redis server.
            port (int): The port of the Redis server.
            password (str): The password for the Redis server.
            embedding (callable): The embedding function to use.
            prefix (str, optional): The prefix for Redis keys. Defaults to "doc".
        """
        import redis

        self.redis_client = redis.Redis(
            host=host,
            port=port,
            password=password
        )
        self.embedding = embedding
        self.prefix = prefix

    def add(self, texts: List[str], metadatas: Union[List[dict], None] = None) -> List[str]:
        """
        Adds texts to the Redis vector store.

        Args:
            texts (List[str]): The texts to be added to the vector store.
            metadatas (List[dict], optional): Metadata for each text. Defaults to None.

        Returns:
            List[str]: List of IDs for the added texts.
        """
        try:
            ids = [str(uuid.uuid4()) for _ in texts]
            embeddings = self.embedding.embed_documents(texts)

            pipe = self.redis_client.pipeline()
            for i, (text, embedding) in enumerate(zip(texts, embeddings)):
                key = f"{self.prefix}:{ids[i]}"
                metadata = json.dumps(metadatas[i] if metadatas else {})

                # Store the text and its embedding
                pipe.hset(key, mapping={
                    "content": text,
                    "embedding": json.dumps(embedding),
                    "metadata": metadata
                })

                pipe.sadd(f"{self.prefix}:keys", key)

            pipe.execute()
            logger.info(f"Added {len(texts)} texts successfully to the vector store.")
            return ids
        except Exception as e:
            logger.error(f"Failed to add texts: {e}")
            raise RuntimeError(f"Can't add texts to the vector store: {e}")

    def _similarity_search_with_score(self, query: str, k: int = 5) -> List[Tuple[Document, float]]:
        """
        Retrieves documents similar to the given query from the Redis vector store.

        Args:
            query (str): The query to retrieve similar documents.
            k (int, optional): The number of top similar documents to retrieve. Defaults to 5.

        Returns:
            List[Tuple[Document, float]]: A list of tuples containing the similar documents and their scores.
        """
        try:
            query_embedding = self.embedding.embed_query(query)

            all_keys = self.redis_client.smembers(f"{self.prefix}:keys")
            results = []

            pipe = self.redis_client.pipeline()
            for key in all_keys:
                pipe.hgetall(key)
            all_docs = pipe.execute()

            for doc_data in all_docs:
                doc_embedding = np.array(json.loads(doc_data[b"embedding"]))

                # Calculate cosine similarity
                similarity = np.dot(query_embedding, doc_embedding) / (
                            np.linalg.norm(query_embedding) * np.linalg.norm(doc_embedding))

                metadata = json.loads(doc_data[b"metadata"].decode()) if b"metadata" in doc_data else {}
                results.append((
                    Document(page_content=doc_data[b"content"].decode(), metadata=metadata),
                    similarity
                ))

            # Sort results by similarity score (descending) and return top k
            return sorted(results, key=lambda x: x[1], reverse=True)[:k]

        except Exception as e:
            logger.error(f"Failed to perform similarity search: {e}")
            raise RuntimeError(f"Can't perform similarity search: {e}")

    def delete(self, ids: List[str]) -> None:
        """
        Deletes documents from the Redis vector store.

        Args:
            ids (List[str]): The list of document IDs to delete.
        """
        try:
            pipe = self.redis_client.pipeline()
            for id in ids:
                pipe.delete(f"{self.prefix}:{id}")
                pipe.srem(f"{self.prefix}:keys", f"{self.prefix}:{id}")
            pipe.execute()
            logger.info(f"Successfully deleted {len(ids)} documents.")
        except Exception as e:
            logger.error(f"Failed to delete documents: {e}")
            raise RuntimeError(f"Can't delete documents from the vector store: {e}")



from tenacity import retry, stop_after_attempt, wait_random_exponential
from loguru import logger
import sys

# Set up logging
logger.remove()  # Remove the default logger
logger.add(sys.stdout,
           format="<green>{level}</green>: <level>{message}</level>",
           level="INFO")

logger.add(sys.stdout,
           format="<red>{level}</red>: <level>{message}</level>",
           level="ERROR")


def format_prompt(context, question):
    """
    Formats the prompt for generating a response.

    Args:
        context (str): The context for the prompt.
        question (str): The question for the prompt.

    Returns:
        str: The formatted prompt.
    """
    return f"Given Context: {context} Give the best full answer amongst the option to question {question}"


class OpenAi:
    def __init__(self, api_key, model, base_url=None):
        """
        Initializes the GPT-3 model with the specified model version and an optional prompt template.

        Args:
            api_key (str): The API key for OpenAI.
            model (str): The GPT-3 model version.
        """
        from openai import OpenAI

        try:
            logger.info(f"Initializing OpenAi with model: {model}")
            self.model = model
            if base_url:
                self.client = OpenAI(api_key=api_key, base_url=base_url)
            else:
                self.client = OpenAI(api_key=api_key)
            logger.info("OpenAi initialized successfully")
        except Exception as e:
            logger.error(f"Error initializing OpenAi: {e}")
            raise

    @retry(wait=wait_random_exponential(min=1, max=20), stop=stop_after_attempt(6))
    def _generate_response(self, messages, max_tokens, temperature, frequency_penalty, presence_penalty, top_p, stream):
        """
        Generates a response from the OpenAI model.

        Args:
            messages (list): The list of messages to send to the model.
            max_tokens (int): The maximum number of tokens in the generated response.
            temperature (float): The sampling temperature.
            frequency_penalty (float): The frequency penalty.
            presence_penalty (float): The presence penalty.
            top_p (float): The top_p parameter for nucleus sampling.
            stream: Indicates if the response should be streamed.

        Returns:
            str: The generated response.
        """
        try:
            logger.info("Generating response")
            response = self.client.chat.completions.create(
                model=self.model,
                messages=messages,
                temperature=temperature,
                max_tokens=max_tokens,
                frequency_penalty=frequency_penalty,
                presence_penalty=presence_penalty,
                top_p=top_p,
                stream=stream
            )

            if stream:
                # If streaming, accumulate the response content
                result = ""
                for chunk in response:
                    content = getattr(chunk.choices[0].delta, 'content', None)
                    if content is not None:
                        result += content
                result = result.strip()
            else:
                # For non-streaming response
                result = response.choices[0].message.content.strip()

            logger.info("Response generated successfully")
            return result

        except Exception as e:
            logger.error(f"Error generating response: {e}")
            raise

    def chat(self, prompt, system_prompt="You are a helpful assistant", max_tokens=None, temperature=0.2,
             frequency_penalty=None, presence_penalty=None, top_p=None, stream=None):
        """
        Public method to interact with the model using chat messages.

        Args:
            prompt (str): The prompt to generate a response for.
            system_prompt (str): The system prompt.
            max_tokens (int, optional): The maximum number of tokens in the generated response. Defaults to None.
            temperature (float, optional): The temperature of the generated response.
            frequency_penalty (float, optional): The frequency penalty.
            presence_penalty (float, optional): The presence penalty.
            top_p (float, optional): The nucleus sampling parameter.
            stream

        Returns:
            str: The generated response.
        """

        messages = [
            {"role": "system", "content": system_prompt},
            {"role": "user", "content": prompt},
        ]
        return self._generate_response(
            messages=messages,
            max_tokens=max_tokens,
            temperature=temperature,
            frequency_penalty=frequency_penalty,
            presence_penalty=presence_penalty,
            top_p=top_p,
            stream=stream
        )

    def answer_question(self, context, question, max_tokens=350, temperature=0.3, frequency_penalty=None,
                        presence_penalty=None, top_p=None,stream=False):
        """
        Public method to generate an answer to a question based on the given context.

        Args:
            context (str): The text to summarize.
            question (str): The question to answer.
            max_tokens (int, optional): The maximum number of tokens in the generated response. Defaults to 350.
            temperature (float, optional): The temperature of the generated response. Defaults to 0.3.
            frequency_penalty (float, optional): The frequency penalty.
            presence_penalty (float, optional): The presence penalty.
            top_p (float, optional): The top_p parameter for nucleus sampling.

        Returns:
            str: The generated answer.
        """
        try:
            logger.info("Answering question")
            prompt = format_prompt(context, question)
            messages = [
                {"role": "system", "content": "You are Question Answering Portal"},
                {"role": "user", "content": prompt},
            ]
            response =  self._generate_response(
                messages=messages,
                max_tokens=max_tokens,
                temperature=temperature,
                frequency_penalty=frequency_penalty,
                presence_penalty=presence_penalty,
                top_p=top_p,
                stream=stream
            )
            return response
        except Exception as e:
            logger.error(f"Error in answer_question: {e}")
            return str(e)

    def get_summary(self, documentation, max_tokens=350, temperature=0.3, frequency_penalty=None,
                    presence_penalty=None, top_p=None):
        """
        Generates a detailed summary of the provided documentation.

        Args:
            documentation (str): The documentation to summarize.
            max_tokens (int, optional): The maximum number of tokens in the generated response. Defaults to 350.
            temperature (float, optional): The temperature of the generated response. Defaults to 0.3.
            frequency_penalty (float, optional): The frequency penalty.
            presence_penalty (float, optional): The presence penalty.
            top_p (float, optional): The top_p parameter for nucleus sampling.

        Returns:
            str: The generated summary.
        """
        try:
            logger.info("Generating summary for documentation")
            prompt = f"You are a helpful assistant. Give a detailed summary of the documentation provided.\n\nDocumentation:\n{documentation}"
            messages = [
                {"role": "system", "content": "You are a helpful assistant"},
                {"role": "user", "content": prompt},
            ]
            return self._generate_response(
                messages=messages,
                max_tokens=max_tokens,
                temperature=temperature,
                frequency_penalty=frequency_penalty,
                presence_penalty=presence_penalty,
                top_p=top_p
            )
        except Exception as e:
            logger.error(f"Error generating summary: {e}")
            return str(e)

    def grade_docs(self, context, question, max_tokens=150, temperature=0.2, frequency_penalty=None,
                   presence_penalty=None, top_p=None):
        """
        Answers a question using an agent-based approach with access to tools.

        Args:
            context (list): The context in which the question is asked.
            question (str): The question to answer.
            max_tokens (int, optional): The maximum number of tokens in the generated response. Defaults to 150.
            temperature (float, optional): The temperature of the generated response.
            frequency_penalty (float, optional): The frequency penalty.
            presence_penalty (float, optional): The presence penalty.
            top_p (float, optional): The top_p parameter for nucleus sampling.

        Returns:
            list: Filtered relevant documents.
        """
        filtered_docs = []
        system_prompt = """
            You are a grader assessing the relevance of a retrieved document to a user question.
            If the document contains any keywords or evidence related to the user question, even if minimal, 
            grade it as relevant. The goal is to filter out only completely erroneous retrievals.
            Give a binary "yes" or "no" score to indicate whether the document is relevant to the question.

            Provide the score with no preamble or explanation.
        """
        for doc in context:
            prompt = f"Here is the retrieved document:\n{doc}\nHere is the user question:\n{question}"
            messages = [
                {"role": "system", "content": system_prompt},
                {"role": "user", "content": prompt},
            ]
            try:
                grade = self._generate_response(
                    messages=messages,
                    max_tokens=max_tokens,
                    temperature=temperature,
                    frequency_penalty=frequency_penalty,
                    presence_penalty=presence_penalty,
                    top_p=top_p
                ).lower()
                if grade == "yes":
                    logger.info("Relevant doc")
                    filtered_docs.append(doc)
                elif grade == "no":
                    logger.info("Not relevant doc")
            except Exception as e:
                logger.error(f"Error grading document: {e}")
        return filtered_docs

    def check_hallucination(self, context, answer, max_tokens=150, temperature=0.2, frequency_penalty=None,
                            presence_penalty=None, top_p=None):
        """
        Checks if an answer is grounded in the provided context.

        Args:
            context (str): The context for checking.
            answer (str): The answer to check.
            max_tokens (int, optional): The maximum number of tokens in the generated response. Defaults to 150.
            temperature (float, optional): The temperature of the generated response. Defaults to 0.2.
            frequency_penalty (float, optional): The frequency penalty.
            presence_penalty (float, optional): The presence penalty.
            top_p (float, optional): The top_p parameter for nucleus sampling.
        Returns:
            str: 'yes' if the answer is grounded, 'no' otherwise.
        """
        system_prompt = """
            You are a grader assessing whether an answer is grounded in / supported by a set of facts.
            Give a binary score 'yes' or 'no' to indicate whether the answer is grounded / supported by the set of facts.
            Provide the score with no preamble or explanation.
        """
        prompt = f"Here are the facts:\n{context}\nHere is the answer:\n{answer}"
        messages = [
            {"role": "system", "content": system_prompt},
            {"role": "user", "content": prompt},
        ]
        try:
            logger.info("Checking hallucination for answer")
            return self._generate_response(
                messages=messages,
                max_tokens=max_tokens,
                temperature=temperature,
                frequency_penalty=frequency_penalty,
                presence_penalty=presence_penalty,
                top_p=top_p
            ).lower()
        except Exception as e:
            logger.error(f"Error checking hallucination: {e}")
            return str(e)



from typing import Any, Dict, List, Optional

from pydantic import Field, root_validator
from indox.core import Embeddings
from loguru import logger
import sys


# Set up logging
logger.remove()  # Remove the default logger
logger.add(sys.stdout,
           format="<green>{level}</green>: <level>{message}</level>",
           level="INFO")

logger.add(sys.stdout,
           format="<red>{level}</red>: <level>{message}</level>",
           level="ERROR")


class ClarifaiEmbeddings(Embeddings):
    """Clarifai embedding models.

    To use, you should have the `clarifai` python package installed, and the
    environment variable `CLARIFAI_PAT` set with your personal access token or pass it
    as a named parameter to the constructor.
    """

    model_url: Optional[str] = None
    model_id: Optional[str] = None
    model_version_id: Optional[str] = None
    app_id: Optional[str] = None
    user_id: Optional[str] = None
    pat: Optional[str] = Field(default=None, exclude=True)
    token: Optional[str] = Field(default=None, exclude=True)
    model: Any = Field(default=None, exclude=True)  #: :meta private:
    api_base: str = "https://api.clarifai.com"

    class Config:
        extra = "forbid"

    @root_validator(pre=True)
    def validate_environment(cls, values: Dict) -> Dict:
        """Validate that we have all required info to access Clarifai platform
        and ensure the Python package exists in the environment."""
        try:
            from clarifai.client.model import Model
        except ImportError:
            raise ImportError(
                "Could not import clarifai python package. "
                "Please install it with `pip install clarifai`."
            )

        required_fields = ['model_id', 'pat']
        for field in required_fields:
            if not values.get(field):
                raise ValueError(f"{field} is required for ClarifaiEmbeddings.")

        values["model"] = Model(
            url=values.get("model_url"),
            app_id=values.get("app_id"),
            user_id=values.get("user_id"),
            model_version=dict(id=values.get("model_version_id")),
            pat=values.get("pat"),
            token=values.get("token"),
            model_id=values.get("model_id"),
            base_url=values.get("api_base"),
        )

        return values

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        """Call out to Clarifai's embedding models.

        Args:
            texts: The list of texts to embed.

        Returns:
            List of embeddings, one for each text.
        """
        if not texts:
            logger.warning("No texts provided for embedding.")
            return []

        from clarifai.client.input import Inputs

        input_obj = Inputs.from_auth_helper(self.model.auth_helper)
        batch_size = 32
        embeddings = []

        try:
            for i in range(0, len(texts), batch_size):
                batch = texts[i: i + batch_size]
                input_batch = [
                    input_obj.get_text_input(input_id=str(id), raw_text=inp)
                    for id, inp in enumerate(batch)
                ]
                predict_response = self.model.predict(input_batch)
                embeddings.extend(
                    [
                        list(output.data.embeddings[0].vector)
                        for output in predict_response.outputs
                    ]
                )
        except (ConnectionError, TimeoutError) as e:
            logger.error(f"Connection or timeout error during prediction: {e}")
        except Exception as e:
            logger.error(f"Prediction failed: {e}")

        return embeddings

    def embed_query(self, text: str) -> List[float]:
        """Call out to Clarifai's embedding models.

        Args:
            text: The text to embed.

        Returns:
            Embeddings for the text.
        """
        try:
            predict_response = self.model.predict_by_bytes(
                bytes(text, "utf-8"), input_type="text"
            )
            embeddings = [
                list(op.data.embeddings[0].vector) for op in predict_response.outputs
            ]
        except (ConnectionError, TimeoutError) as e:
            logger.error(f"Connection or timeout error during prediction: {e}")
            return []
        except Exception as e:
            logger.error(f"Prediction failed: {e}")
            return []

        return embeddings[0] if embeddings else []



from indox.data_connectors.utils import Document
from typing import Any, List, Optional


class TwitterTweetReader:
    """Twitter tweets reader.

    This class reads tweets from a specified user's Twitter handle. To use this
    class, you must have access to the Twitter API. For more information on how to
    obtain access, refer to the following guide:
    'https://developer.twitter.com/en/docs/twitter-api/getting-started/getting-access-to-the-twitter-api'.

    Args:
        bearer_token (str): Bearer token obtained from the Twitter API.
        num_tweets (Optional[int]): The number of tweets to fetch for each user's Twitter
            handle. The default is 100 tweets.
    """

    is_remote: bool = True
    bearer_token: str
    num_tweets: Optional[int]

    def __init__(
            self,
            bearer_token: str,
            num_tweets: Optional[int] = 100,
    ) -> None:
        """Initialize TwitterTweetReader with the given parameters.

        Args:
            bearer_token (str): Bearer token obtained from the Twitter API.
            num_tweets (Optional[int], optional): The number of tweets to fetch for each
                user's Twitter handle. Defaults to 100.
        """
        self.bearer_token = bearer_token
        self.num_tweets = num_tweets

    @classmethod
    def class_name(cls) -> str:
        """Returns the class name as a string.

        Returns:
            str: The name of the class, "TwitterTweetReader".
        """
        return "TwitterTweetReader"

    def load_data(
            self,
            twitterhandles: List[str],
            num_tweets: Optional[int] = None,
            **load_kwargs: Any
    ) -> List[Document] | Document:
        """Load tweets for specified Twitter handles.

        Args:
            twitterhandles (List[str]): List of Twitter handles to read tweets from.
            num_tweets (Optional[int], optional): The number of tweets to fetch for each handle.
                If not provided, the instance's `num_tweets` will be used.

        Returns:
            List[Document]: A list of Document instances containing the fetched tweets
            and their associated metadata.

        Raises:
            ImportError: If the `tweepy` package is not installed.
            tweepy.TweepError: If there is an error with the Twitter API request.
            ValueError: If a Twitter handle is invalid or the user does not exist.
        """
        try:
            import tweepy
        except ImportError:
            raise ImportError(
                "`tweepy` package not found. Please install it using `pip install tweepy`."
            )

        try:
            client = tweepy.Client(bearer_token=self.bearer_token)
            documents = []
            for username in twitterhandles:
                user = client.get_user(username=username)
                if not user.data:
                    raise ValueError(f"User '{username}' not found or is invalid.")

                tweets = client.get_users_tweets(
                    user.data.id, max_results=num_tweets or self.num_tweets
                )

                if not tweets.data:
                    raise ValueError(f"No tweets found for user '{username}'.")

                response = "\n".join(tweet.text for tweet in tweets.data)
                metadata = {
                    "username": username,
                    "user_id": user.data.id,
                    "num_tweets": len(tweets.data),
                }

                documents.append(Document(source="Twitter", content=response, metadata=metadata))
            if len(documents) == 1:
                return documents[0]
            return documents

        except tweepy.TweepyException as e:
            raise tweepy.TweepyException(f"An error occurred with the Twitter API: {e}")

    def load_content(
            self,
            twitterhandles: List[str],
            num_tweets: Optional[int] = None,
            **load_kwargs: Any
    ) -> List[str] | str:
        """Load tweets for specified Twitter handles.

        Args:
            twitterhandles (List[str]): List of Twitter handles to read tweets from.
            num_tweets (Optional[int], optional): The number of tweets to fetch for each handle.
                If not provided, the instance's `num_tweets` will be used.

        Returns:
            List[str]: A list of strings, each containing concatenated tweets from a user.

        Raises:
            ImportError: If the `tweepy` package is not installed.
            tweepy.TweepError: If there is an error with the Twitter API request.
            ValueError: If a Twitter handle is invalid or the user does not exist.
        """
        try:
            import tweepy
        except ImportError:
            raise ImportError(
                "`tweepy` package not found. Please install it using `pip install tweepy`."
            )

        try:
            client = tweepy.Client(bearer_token=self.bearer_token)
            tweets_contents = []
            for username in twitterhandles:
                user = client.get_user(username=username)
                if not user.data:
                    raise ValueError(f"User '{username}' not found or is invalid.")

                tweets = client.get_users_tweets(
                    user.data.id, max_results=num_tweets or self.num_tweets
                )

                if not tweets.data:
                    raise ValueError(f"No tweets found for user '{username}'.")

                response = "\n".join(tweet.text for tweet in tweets.data)
                tweets_contents.append(response)
            if len(tweets_contents) == 1:
                return tweets_contents[0]
            return tweets_contents

        except tweepy.TweepyException as e:
            raise tweepy.TweepyException(f"An error occurred with the Twitter API: {e}")



import requests
from loguru import logger
import sys
from indox.core import BaseLLM

# Set up logging
logger.remove()  # Remove the default logger
logger.add(sys.stdout,
           format="<green>{level}</green>: <level>{message}</level>",
           level="INFO")

logger.add(sys.stdout,
           format="<red>{level}</red>: <level>{message}</level>",
           level="ERROR")


class IndoxApi:
    api_key: str

    def __init__(self, api_key, prompt_template=""):
        """
        Initializes the IndoxApi with the specified API key and an optional prompt template.

        Args:
            api_key (str): The API key for Indox API.
            prompt_template (str, optional): The template for the prompt. Defaults to None.
        """
        self.api_key = api_key
        self.prompt_template = prompt_template or "Context: {context}\nQuestion: {question}\nAnswer:"

    def _send_request(self, system_prompt, user_prompt, max_tokens, temperature, stream, model,
                      presence_penalty, frequency_penalty, top_p):
        url = 'http://5.78.55.161/api/chat_completion/generate/'
        headers = {
            'accept': '*/*',
            "Authorization": f"Bearer {self.api_key}",
            'Content-Type': 'application/json',
        }

        data = {
            "frequency_penalty": frequency_penalty,
            "max_tokens": max_tokens,
            "messages": [
                {
                    "content": system_prompt,
                    "role": "system"
                },
                {
                    "content": user_prompt,
                    "role": "user"
                }
            ],
            "model": model,
            "presence_penalty": presence_penalty,
            "stream": stream,
            "temperature": temperature,
            "top_p": top_p
        }

        response = requests.post(url, headers=headers, json=data)
        if response.status_code == 200:
            answer_data = response.json()
            generated_text = answer_data.get("text_message", "")
            return generated_text
        else:
            raise Exception(f"Error From Indox API: {response.status_code}, {response.text}")

    def _attempt_answer_question(self, context, question, max_tokens, temperature, stream, model,
                                 presence_penalty, frequency_penalty, top_p):
        """
        Generates an answer to a question based on the given context using the Indox API.

        Args:
            context (str): The text to base the answer on.
            question (str): The question to be answered.

        Returns:
            str: The generated answer.
        """
        system_prompt = "You are a helpful assistant."
        user_prompt = self.prompt_template.format(context=context, question=question, )
        return self._send_request(system_prompt, user_prompt, max_tokens=max_tokens,
                                  temperature=temperature, stream=stream,
                                  model=model, presence_penalty=presence_penalty,
                                  frequency_penalty=frequency_penalty,
                                  top_p=top_p)

    def answer_question(self, context, question, max_tokens=350, temperature=0.3, stream=True,
                        model="gpt-4o-mini", presence_penalty=0, frequency_penalty=0, top_p=1):
        """
        Answer a question based on the given context using the Indox API.

        Returns:
            str: The generated answer.
        """
        try:
            # prompt_template = prompt_template or self.prompt_template
            return self._attempt_answer_question(context, question, max_tokens=max_tokens,
                                                 temperature=temperature, stream=stream,
                                                 model=model, presence_penalty=presence_penalty,
                                                 frequency_penalty=frequency_penalty,
                                                 top_p=top_p)
        except Exception as e:
            logger.error(e)
            return str(e)

    def get_summary(self, documentation, max_tokens=350, temperature=0.3, stream=True,
                    model="gpt-4o-mini", presence_penalty=0, frequency_penalty=0, top_p=1):
        """
        Generates a detailed summary of the provided documentation.


        Returns:
            str: The generated summary.
        """
        try:
            system_prompt = "You are a helpful assistant."
            user_prompt = f"You are a helpful assistant. Give a detailed summary of the documentation provided.\n\nDocumentation:\n {documentation}"
            return self._send_request(system_prompt=system_prompt, user_prompt=user_prompt,
                                      max_tokens=max_tokens,
                                      temperature=temperature, stream=stream,
                                      model=model, presence_penalty=presence_penalty,
                                      frequency_penalty=frequency_penalty,
                                      top_p=top_p)
        except Exception as e:
            logger.error(e)
            return str(e)

    def grade_docs(self, context, question, max_tokens=350, temperature=0.3, stream=True,
                   model="gpt-4o-mini", presence_penalty=0, frequency_penalty=0, top_p=1):
        """
        Answers a question using an agent-based approach with access to tools.

        Returns:
            list: Filtered list of relevant documents.
        """
        filtered_docs = []
        try:
            system_prompt = """
            You are a grader assessing relevance of a retrieved document to a user question. If the document contains keywords related to the user question, grade it as relevant. It does not need to be a stringent test. The goal is to filter out erroneous retrievals.
            Give a binary score 'yes' or 'no' score to indicate whether the document is relevant to the question.
            Provide the score with no preamble or explanation.
            """
            for i in range(len(context)):
                user_prompt = f"""
                Here is the retrieved document:
                {context[i]}
                Here is the user question:
                {question}
                """
                response = self._send_request(system_prompt=system_prompt, user_prompt=user_prompt,
                                              max_tokens=max_tokens,
                                              temperature=temperature, stream=stream,
                                              model=model, presence_penalty=presence_penalty,
                                              frequency_penalty=frequency_penalty,
                                              top_p=top_p)
                if response.lower() == "yes":
                    logger.info("Relevant doc")
                    filtered_docs.append(context[i])
                elif response.lower() == "no":
                    logger.info("Not Relevant doc")
            return filtered_docs
        except Exception as e:
            logger.error(f"Error generating agent answer: {e}")
            return str(e)

    def check_hallucination(self, context, answer, max_tokens=350, temperature=0.3, stream=True,
                            model="gpt-4o-mini", presence_penalty=0, frequency_penalty=0, top_p=1):
        """
        Checks if an answer is grounded in the provided context.

        Args:
            context (str): The text to base the answer on.
            answer (str): The answer to check for hallucination.

        Returns:
            str: 'yes' if the answer is grounded, 'no' otherwise.
        """
        try:
            system_prompt = """
            You are a grader assessing whether an answer is grounded in / supported by a set of facts. Give a binary score 'yes' or 'no' score to indicate whether the answer is grounded / supported by a set of facts. Provide score with no preamble or explanation.
            """
            user_prompt = f"""
            Here are the facts:
            \n -------- \n
            {context}
            \n -------- \n
            Here is the answer: {answer}
            """
            response = self._send_request(system_prompt=system_prompt, user_prompt=user_prompt, max_tokens=max_tokens,
                                          temperature=temperature, stream=stream,
                                          model=model, presence_penalty=presence_penalty,
                                          frequency_penalty=frequency_penalty,
                                          top_p=top_p)
            return response
        except Exception as e:
            logger.error(f"Error generating agent answer: {e}")
            return str(e)

    # def _assistant(self, system_prompt, user_prompt, role, history):
    #     try:
    #         response = self._send_request(system_prompt=system_prompt, user_prompt=user_prompt, role=role,
    #                                       history=history)
    #         return response
    #     except Exception as e:
    #         logger.error(f"Error generating agent answer: {e}")
    #         return str(e)

    def chat(self, prompt, system_prompt="You are a helpful assistant", max_tokens=16384, temperature=0.3, stream=True,
             model="gpt-4o-mini", presence_penalty=0, frequency_penalty=0, top_p=1):
        return self._send_request(system_prompt=system_prompt, user_prompt=prompt, max_tokens=max_tokens,
                                  temperature=temperature, stream=stream,
                                  model=model, presence_penalty=presence_penalty, frequency_penalty=frequency_penalty,
                                  top_p=top_p)



import os
import io
import pandas as pd
import nbformat
import json

class GoogleDrive:
    SCOPES = ['https://www.googleapis.com/auth/drive.readonly']

    def __init__(self, creds_file='tokenGoogleDrive.json', credentials_json='credentials.json'):
        """
        Initialize the GoogleDrive object and authenticate using OAuth2.

        Parameters:
        - creds_file (str): The path to the file where the credentials are stored.
        - credentials_json (str): The path to the JSON file containing the client secrets.

        Returns:
        - None
        """

        self.creds_file = creds_file
        self.credentials_json = credentials_json
        self.creds = self._authenticate()

    def _authenticate(self):
        """
        Authenticate the user and refresh credentials if necessary.

        Returns:
        - Credentials: The authenticated Google API credentials.
        """
        from google.auth.transport.requests import Request

        from google.oauth2.credentials import Credentials
        from google_auth_oauthlib.flow import InstalledAppFlow


        creds = None
        if os.path.exists(self.creds_file):
            print("Loading credentials from file.")
            creds = Credentials.from_authorized_user_file(self.creds_file, self.SCOPES)
        if not creds or not creds.valid:
            if creds and creds.expired and creds.refresh_token:
                print("Refreshing expired credentials.")
                try:
                    creds.refresh(Request())
                except Exception as e:
                    print(f"Failed to refresh credentials: {e}")
                    raise
            else:
                print("Authenticating new credentials.")
                try:
                    flow = InstalledAppFlow.from_client_secrets_file(self.credentials_json, self.SCOPES)
                    creds = flow.run_local_server(port=0)
                except Exception as e:
                    print(f"Failed to authenticate: {e}")
                    raise
            try:
                with open(self.creds_file, 'w') as token:
                    token.write(creds.to_json())
                print("Credentials saved to token file.")
            except Exception as e:
                print(f"Failed to save credentials: {e}")
                raise
        return creds

    def read(self, file_id):
        """
        Read and print content from a file in Google Drive based on its MIME type.

        Parameters:
        - file_id (str): The ID of the file in Google Drive to be read.

        Returns:
        - None
        """
        import xml.etree.ElementTree as ET
        from bs4 import BeautifulSoup
        from googleapiclient.discovery import build
        import PyPDF2
        from pptx import Presentation
        from docx import Document

        try:
            service = build('drive', 'v3', credentials=self.creds)
            file = service.files().get(fileId=file_id, fields='id, name, mimeType, webViewLink').execute()

            file_name = file.get('name')
            file_mime_type = file.get('mimeType')

            print(f"File Name: {file_name}")
            print(f"MIME Type: {file_mime_type}")
            print(f"Web View Link: {file.get('webViewLink')}")

            request = service.files().get_media(fileId=file_id)
            file_content = request.execute()

            if file_mime_type == 'text/plain':  # .txt, .py
                print("File Content:\n")
                print(file_content.decode('utf-8'))

            elif file_name.endswith('.ipynb'):  # .ipynb
                nb = nbformat.reads(file_content.decode('utf-8'), as_version=4)
                print("Jupyter Notebook Content:\n")
                for cell in nb.cells:
                    if cell.cell_type == 'code':
                        print(f"Code:\n{cell.source}\n")

            elif file_mime_type == 'application/pdf':  # .pdf
                pdf_reader = PyPDF2.PdfReader(io.BytesIO(file_content))
                print("PDF Content:\n")
                for page in pdf_reader.pages:
                    print(page.extract_text() or "No text found on this page.")

            elif file_mime_type == 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet':  # .xlsx
                df = pd.read_excel(io.BytesIO(file_content))
                print("Excel Content:\n")
                print(df)

            elif file_mime_type == 'text/csv':  # .csv
                df = pd.read_csv(io.BytesIO(file_content))
                print("CSV Content:\n")
                print(df)

            elif file_mime_type == 'application/vnd.openxmlformats-officedocument.presentationml.presentation':  # PowerPoint
                ppt = Presentation(io.BytesIO(file_content))
                print("PowerPoint Content:\n")
                for slide in ppt.slides:
                    for shape in slide.shapes:
                        if hasattr(shape, "text"):
                            print(shape.text)

            elif file_mime_type == 'application/vnd.openxmlformats-officedocument.wordprocessingml.document':  # .docx
                doc = Document(io.BytesIO(file_content))
                print("Word Document Content:\n")
                for paragraph in doc.paragraphs:
                    print(paragraph.text)

            elif file_mime_type == 'application/json':  # .json
                json_content = json.loads(file_content.decode('utf-8'))
                print("JSON Content:\n")
                print(json.dumps(json_content, indent=4))

            elif file_mime_type == 'application/xml':  # .xml
                tree = ET.ElementTree(ET.fromstring(file_content.decode('utf-8')))
                root = tree.getroot()
                print("XML Content:\n")
                ET.dump(root)

            elif file_mime_type == 'text/html':  # .html
                soup = BeautifulSoup(file_content, 'html.parser')
                print("HTML Content:\n")
                print(soup.prettify())

            else:
                print(f"Unsupported MIME type: {file_mime_type}")

        except Exception as err:
            print(f"An error occurred: {err}")
        finally:
            if os.path.exists(self.creds_file):
                os.remove(self.creds_file)





import abc
from pydantic import BaseModel


class VectorStore(BaseModel, abc.ABC):
    """
    Abstract base class for vector stores.
    """

    @abc.abstractmethod
    def add_document(self, docs):
        pass

    @abc.abstractmethod
    def retrieve(self, query: str, top_k: int = 5):
        pass

    @abc.abstractmethod
    def get_all_documents(self):
        pass

    class Config:
        arbitrary_types_allowed = True



import json
from typing import Any, Dict, List, Optional, Tuple, Union, Callable, Literal, Iterable
import uuid
from indox.core import Embeddings, VectorStore, Document
from loguru import logger
import sys


# Set up logging
logger.remove()  # Remove the default logger
logger.add(sys.stdout,
           format="<green>{level}</green>: <level>{message}</level>",
           level="INFO")

logger.add(sys.stdout,
           format="<red>{level}</red>: <level>{message}</level>",
           level="ERROR")



class Milvus:
    """
    A wrapper class for interacting with the Milvus vector database.
    """
    def __init__(self, collection_name, embedding_model):
        """
        Initialize the MilvusWrapper with collection name, embedding model, and QA model.

        Args:
            collection_name (str): The name of the collection in Milvus.
            embedding_model (object): An object with methods `embed_query` and `embed_documents`.
        """
        from pymilvus import MilvusClient, MilvusException

        self.collection_name = collection_name
        self.embedding_model = embedding_model
        self.embedding_dim = None
        self.milvus_client = MilvusClient(host='127.0.0.1', port='19530')


    def _embed_query(self, query: str) -> List[float]:
        """
        Embed the query into a vector.

        Args:
            query (str): The query text to embed.

        Returns:
            List[float]: The embedding vector for the query.
        """
        return self.embedding_model.embed_query(query)

    def _similarity_search_with_score(self, query: str, k: int = 3) -> List[Tuple[Document, float]]:
        """
        Return docs most similar to the query.

        Args:
            query (str): Text to look up documents similar to.
            k (int, optional): Number of Documents to return. Defaults to 3.

        Returns:
            List[Tuple[Document, float]]: List of documents most similar to the query text with
            L2 distance in float. Lower score represents more similarity.
        """
        search_res = self.milvus_client.search(
            collection_name=self.collection_name,
            data=[self.emb_text(query)],
            limit=k,
            search_params={"metric_type": "IP", "params": {}},
            output_fields=["text"],
        )

        # Creating Document objects for the results
        documents_with_scores = [
            (Document(page_content=res["entity"]["text"]), res["distance"])
            for res in search_res[0]
        ]
        return documents_with_scores
    def _similarity_search(
        self,
        query: str,
        k: int = 4,
        param: Optional[dict] = None,
        expr: Optional[str] = None,
        timeout: Optional[float] = None,
        **kwargs: Any,
    ) -> List[Document]:
        """Perform a similarity search against the query string.

        Args:
            query (str): The text to search.
            k (int, optional): How many results to return. Defaults to 4.
            param (dict, optional): The search params for the index type.
                Defaults to None.
            expr (str, optional): Filtering expression. Defaults to None.
            timeout (int, optional): How long to wait before timeout error.
                Defaults to None.
            kwargs: Collection.search() keyword arguments.

        Returns:
            List[Document]: Document results for search.
        """
        if self.col is None:
            logger.debug("No existing collection to search.")
            return []
        res = self._similarity_search_with_score(
            query=query, k=k
        )
        return [doc for doc, _ in res]

    def _process_question(self, question: str):
        """
        Process a question by retrieving relevant documents and printing them.

        Args:
            question (str): The question to process.
        """
        retrieved_lines_with_distances = self._similarity_search_with_score(question, k=5)
        # Convert Document objects to dictionaries
        context = "\n".join([self._to_dict(doc)['page_content'] for doc, _ in retrieved_lines_with_distances])
        # answer = self.generate_answer(context, question)
        # print(f"Answer: {answer}")
        print(json.dumps(
            [{"document": self._to_dict(doc), "score": score} for doc, score in retrieved_lines_with_distances],
            indent=4
        ))

    # def store_in_vectorstore(self, docs: List[Document]):
    #     """
    #     Store documents in the Milvus vector store.
    #
    #     Args:
    #         docs (List[Document]): List of Document objects to store.
    #     """
    #     text_lines = [doc.page_content for doc in docs]
    #     self.insert_data(text_lines)

    def emb_text(self, text: str) -> List[float]:
        """
        Get the embedding vector for the provided text.

        Args:
            text (str): The text to embed.

        Returns:
            List[float]: The embedding vector for the text.
        """
        embedding = self.embedding_model.embed_documents([text])[0]
        if self.embedding_dim is None:
            self.embedding_dim = len(embedding)
        return embedding

    def load_text_from_file(self, file_path: str) -> List[str]:
        """
        Load text lines from a file.

        Args:
            file_path (str): Path to the text file.

        Returns:
            List[str]: List of text lines read from the file.
        """
        with open(file_path, "r") as file:
            text_lines = file.readlines()
        return text_lines

    def insert_data(self, text_lines: List[str]):
        """
        Insert data into the Milvus vector store.

        Args:
            text_lines (List[str]): List of text lines to insert.
        """
        from tqdm import tqdm

        data = [{"id": i, "vector": self.emb_text(line), "text": line} for i, line in
                enumerate(tqdm(text_lines, desc="Creating embeddings"))]
        self.milvus_client.insert(collection_name=self.collection_name, data=data)

    def _add(
            self,
            texts: Iterable[str],
            embeddings: Iterable[List[float]],
            metadatas: Optional[Iterable[dict]] = None,
            ids: Optional[List[str]] = None,
    ) -> List[str]:
        """
        Add documents with embeddings to the vector store.

        Args:
            texts (Iterable[str]): Texts to add.
            embeddings (Iterable[List[float]]): Corresponding embeddings for the texts.
            metadatas (Optional[Iterable[dict]]): Optional metadata for the documents.
            ids (Optional[List[str]]): Optional IDs for the documents.

        Returns:
            List[str]: List of document IDs.
        """
        _metadatas = metadatas or ({} for _ in texts)
        documents = [
            Document(page_content=t, metadata=m) for t, m in zip(texts, _metadatas)
        ]

        if ids and len(ids) != len(set(ids)):
            raise ValueError("Duplicate ids found in the ids list.")

        ids = ids or [str(uuid.uuid4()) for _ in texts]
        for id_, doc, embedding in zip(ids, documents, embeddings):
            self.milvus_client.add(embedding=embedding, metadata=doc.metadata)
        return ids

    def add(self, docs):
        """
        Adds documents to the Milvus vector store.

        Args:
            docs: The documents to be added to the vector store.
        """
        try:
            if isinstance(docs[0], Document):
                self._add_documents(documents=docs)
            else:
                self._add_texts(texts=docs)
            logger.info("Document added successfully to the vector store.")
        except Exception as e:
            logger.error(f"Failed to add document: {e}")
            raise RuntimeError(f"Can't add document to the vector store: {e}")

    def _add_texts(
            self,
            texts: Iterable[str],
            metadatas: Optional[List[dict]] = None,
            ids: Optional[List[str]] = None,
            **kwargs: Any,
    ) -> List[str]:
        """
        Add texts with embeddings to the vector store.

        Args:
            texts (Iterable[str]): Texts to add.
            metadatas (Optional[List[dict]]): Optional metadata for the documents.
            ids (Optional[List[str]]): Optional IDs for the documents.

        Returns:
            List[str]: List of document IDs.
        """
        texts = list(texts)
        embeddings = self.embedding_model.embed_documents(texts)
        return self._add(texts, embeddings, metadatas=metadatas, ids=ids)

    def add_embeddings(
            self,
            text_embeddings: Iterable[Tuple[str, List[float]]],
            metadatas: Optional[List[dict]] = None,
            ids: Optional[List[str]] = None,
            **kwargs: Any,
    ) -> List[str]:
        """
        Add embeddings with corresponding texts to the vector store.

        Args:
            text_embeddings (Iterable[Tuple[str, List[float]]]): Iterable of text and embedding pairs.
            metadatas (Optional[List[dict]]): Optional metadata for the documents.
            ids (Optional[List[str]]): Optional IDs for the documents.

        Returns:
            List[str]: List of document IDs.
        """
        texts, embeddings = zip(*text_embeddings)
        return self.add(texts, embeddings, metadatas=metadatas, ids=ids)

    def _add_documents(self, documents: List[Document], **kwargs: Any) -> List[str]:
        """Run more documents through the embeddings and add to the vectorstore.

        Args:
            documents (List[Document]: Documents to add to the vectorstore.

        Returns:
            List[str]: List of IDs of the added texts.
        """
        texts = [doc.page_content for doc in documents]
        metadatas = [doc.metadata for doc in documents]
        return self._add_texts(texts, metadatas, **kwargs)

    def _to_dict(self, document: Document) -> Dict[str, Any]:
        """
        Convert a Document to a dictionary.

        Args:
            document (Document): Document object to convert.

        Returns:
            dict: Dictionary containing the page content and metadata.
        """
        return {
            "page_content": document.page_content,
            "metadata": document.metadata
        }



import os
import requests
from typing import List, Optional


class AI21SemanticTextSplitter:
    """
    A class for splitting text into semantic chunks using the AI21 API.

    This class provides functionality to split text into meaningful segments
    using AI21's semantic text segmentation API. It can also merge these
    segments into larger chunks if a chunk size is specified.
    """

    def __init__(
            self,
            chunk_size: int = 400,
            chunk_overlap: int = 50,
            api_key: Optional[str] = None,
            api_host: str = "https://api.ai21.com/studio/v1",
            timeout_sec: Optional[float] = None,
            num_retries: int = 3,
    ):
        """
        Initialize the AI21SemanticTextSplitter.

        Args:
            chunk_size (int): The maximum size of each chunk. If 0, no merging is done.
            chunk_overlap (int): The number of characters to overlap between chunks.
            api_key (Optional[str]): The AI21 API key. If not provided, it will be read from the AI21_API_KEY environment variable.
            api_host (str): The base URL for the AI21 API.
            timeout_sec (Optional[float]): The timeout for API requests in seconds.
            num_retries (int): The number of times to retry failed API calls.

        Raises:
            ValueError: If no API key is provided or found in the environment.
        """
        self.chunk_size = chunk_size
        self.chunk_overlap = chunk_overlap
        self.api_key = api_key or os.environ.get("AI21_API_KEY")
        self.api_host = api_host
        self.timeout_sec = timeout_sec
        self.num_retries = num_retries

        if not self.api_key:
            raise ValueError("AI21_API_KEY must be provided or set as an environment variable.")

    def split_text(self, text: str) -> List[str]:
        """
        Split the input text into semantic segments.

        This method calls the AI21 API to perform semantic segmentation on the input text.
        If a chunk_size is specified, it also merges the segments into larger chunks.

        Args:
            text (str): The input text to be split.

        Returns:
            List[str]: A list of text segments or chunks.
        """
        segments = self._call_segmentation_api(text)

        if self.chunk_size > 0:
            return self._merge_splits(segments)
        return segments

    def _call_segmentation_api(self, text: str) -> List[str]:
        """
        Call the AI21 segmentation API to split the text.

        This method makes a POST request to the AI21 API to perform semantic segmentation.
        It includes retry logic in case of failed API calls.

        Args:
            text (str): The input text to be segmented.

        Returns:
            List[str]: A list of segmented text pieces.

        Raises:
            Exception: If all API call attempts fail.
        """
        url = f"{self.api_host}/segmentation"
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json"
        }
        data = {
            "source": text,
            "sourceType": "TEXT"
        }

        for _ in range(self.num_retries):
            try:
                response = requests.post(url, json=data, headers=headers, timeout=self.timeout_sec)
                response.raise_for_status()
                result = response.json()
                return [segment['segmentText'] for segment in result['segments']]
            except requests.RequestException as e:
                print(f"API call failed: {e}")

        raise Exception("Failed to call AI21 API after multiple retries")

    def _merge_splits(self, splits: List[str]) -> List[str]:
        """
        Merge the split segments into larger chunks.

        This method combines the split segments into chunks of size up to chunk_size.

        Args:
            splits (List[str]): The list of text segments to be merged.

        Returns:
            List[str]: A list of merged text chunks.
        """
        chunks = []
        current_chunk = ""

        for split in splits:
            if len(current_chunk) + len(split) > self.chunk_size:
                if current_chunk:
                    chunks.append(current_chunk)
                    current_chunk = ""
            current_chunk += split

        if current_chunk:
            chunks.append(current_chunk)

        return chunks



import pandas as pd
from typing import List, Dict, Any, Tuple
from indox.core import Document
import uuid


def extract_title(text: str) -> str:
    """
    Extract the title from the text.

    Args:
        text (str): The input text.

    Returns:
        str: Title extracted from the text, which is the first sentence.
    """
    try:
        if '.' in text:
            title = text.split('.')[0] + '.'
        else:
            title = "Not mentioned"
        return title
    except Exception as e:
        print(f"Error extracting title: {e}")
        return "Not mentioned"


def generate_id() -> str:
    """
    Generate a unique identifier using UUID.

    Returns:
        str: A unique identifier in string format.
    """
    return str(uuid.uuid4())


def process_docs(input_list: List[str], embedding_function: Any) -> List[Dict[str, Any]]:
    """
    Process the list of input texts, extract titles, and generate embeddings.

    Args:
        input_list (List[str]): List of input documents in string format.
        embedding_function (Any): Function to generate embeddings.

    Returns:
        List[Dict[str, Any]]: Processed document objects including embeddings, titles, and ids.
    """
    docs = []
    try:
        for text in input_list:
            doc_id = generate_id()
            embedding = embedding_function.embed_query(text)

            print(f"Embedding shape for document {doc_id}: {len(embedding)}")

            output_dict = {
                "id": doc_id,
                "fields": {
                    "title": extract_title(text),
                    "body": text,
                    "embedding": embedding,
                    "id": doc_id
                }
            }
            docs.append(output_dict)
    except Exception as e:
        print(f"Error processing documents: {e}")
    return docs


class VESPA:

    def __init__(self, app_name: str, embedding_function: Any):
        """
        Initialize the Vespa application with schema and embedding functionality.

        Args:
            app_name (str): Name of the Vespa application.
            embedding_function (Any): Embedding function to handle text embeddings.
        """
        try:
            from vespa.package import (
                ApplicationPackage,
                Field,
                Schema,
                Document as VespaDocument,
                HNSW,
                RankProfile,
                FieldSet,
                GlobalPhaseRanking,
                Function,
            )
            from vespa.deployment import VespaDocker
            from vespa.io import VespaResponse

            self.app_name = app_name
            self.embedding_function = embedding_function

            # Get the embedding for the test sentence to determine its dimensions
            test_sentence = "This is a test."
            test_embedding = self.embedding_function.embed_query(test_sentence)
            self.embedding_dim = len(test_embedding)

            print(f"Detected embedding dimension: {self.embedding_dim}")

            # Update schema to use dynamic embedding dimension
            self.package = ApplicationPackage(
                name=self.app_name,
                schema=[
                    Schema(
                        name="doc",
                        document=VespaDocument(
                            fields=[
                                Field(name="id", type="string", indexing=["summary"]),
                                Field(
                                    name="title",
                                    type="string",
                                    indexing=["index", "summary"],
                                    index="enable-bm25",
                                ),
                                Field(
                                    name="body",
                                    type="string",
                                    indexing=["index", "summary"],
                                    index="enable-bm25",
                                    bolding=True,
                                ),
                                Field(
                                    name="embedding",
                                    type=f"tensor<float>(x[{self.embedding_dim}])",
                                    indexing=["index", "attribute"],
                                    ann=HNSW(distance_metric="angular"),
                                ),
                            ]
                        ),
                        fieldsets=[FieldSet(name="default", fields=["title", "body"])],
                        rank_profiles=[
                            RankProfile(
                                name="bm25",
                                functions=[
                                    Function(name="bm25sum", expression="bm25(title) + bm25(body)")
                                ],
                                first_phase="bm25sum",
                            ),
                            RankProfile(
                                name="semantic",
                                inputs=[
                                    ("query(q)", f"tensor<float>(x[{self.embedding_dim}])"),
                                ],
                                first_phase="closeness(field, embedding)",
                            ),
                            RankProfile(
                                name="fusion",
                                inherits="bm25",
                                inputs=[("query(q)", f"tensor<float>(x[{self.embedding_dim}])")],
                                first_phase="closeness(field, embedding)",
                                global_phase=GlobalPhaseRanking(
                                    expression="reciprocal_rank_fusion(bm25sum, closeness(field, embedding))",
                                    rerank_count=1000,
                                ),
                            ),
                        ],
                    )
                ]
            )
            self.vespa_docker = VespaDocker()
            self.app = None
        except Exception as e:
            print(f"Error initializing VESPA application: {e}")

    def deploy(self):
        """
        Deploy the Vespa application using Docker.

        Raises:
            Exception: If deployment fails.
        """
        try:
            self.app = self.vespa_docker.deploy(application_package=self.package)
        except Exception as e:
            print(f"Error deploying VESPA application: {e}")
            raise

    def add(self, docs: List[Dict[str, Any]]):
        """
        Add documents to Vespa schema.

        Args:
            docs (List[Dict[str, Any]]): List of documents to add to the Vespa schema.
        """
        try:
            from vespa.io import VespaResponse
            docs = process_docs(docs, self.embedding_function)

            def callback(response: VespaResponse, id: str):
                if not response.is_successful():
                    print(f"Error when feeding document {id}: {response.get_json()}")

            self.app.feed_iterable(docs, schema="doc", namespace=self.app_name, callback=callback)
        except Exception as e:
            print(f"Error adding documents to VESPA: {e}")

    def _get_relevant_documents(self, query: str) -> List[Dict[str, Any]]:
        """
        Retrieve relevant documents based on the query.

        Args:
            query (str): Input query text.

        Returns:
            List[Dict[str, Any]]: List of relevant documents.

        Raises:
            ValueError: If the query fails.
        """
        try:
            embedded_query = self.embedding_function.embed_query(query)
            response = self.app.query(
                yql="select * from sources * where ({targetHits:1000}nearestNeighbor(embedding,q)) limit 5",
                query=query,
                ranking="semantic",
                body={"input.query(q)": embedded_query},
            )
            if not response.is_successful():
                raise ValueError(
                    f"Query failed with status code {response.status_code}, url={response.url} response={response.get_json()}"
                )
            return self._parse_response(response.get_json())
        except Exception as e:
            print(f"Error retrieving relevant documents: {e}")
            raise

    def _parse_response(self, response: Dict[str, Any]) -> List[Document]:
        """
        Parse the response from Vespa into a list of Document objects.

        Args:
            response (Dict[str, Any]): The JSON response from Vespa.

        Returns:
            List[Document]: Parsed list of document objects.
        """
        try:
            documents: List[Document] = []
            hits = response.get('root', {}).get('children', [])
            for hit in hits:
                fields = hit.get('fields', {})
                documents.append(Document(
                    page_content=fields.get('body', ''),
                    relevance=hit.get('relevance', 0)
                ))
            return documents
        except Exception as e:
            print(f"Error parsing response: {e}")
            return []

    def display_hits_as_df(self, documents: List[Document]) -> pd.DataFrame:
        """
        Display document hits as a pandas DataFrame.

        Args:
            documents (List[Document]): List of Document objects.

        Returns:
            pd.DataFrame: DataFrame containing document bodies and relevance scores.
        """
        try:
            hits_data = []
            for doc in documents:
                hit_data = {
                    "body": doc.page_content,
                    "score": doc.metadata.get("relevance", 0)
                }
                hits_data.append(hit_data)

            df = pd.DataFrame(hits_data)
            return df
        except Exception as e:
            print(f"Error displaying hits as DataFrame: {e}")
            return pd.DataFrame()

    def similarity_search_with_score(self, query: str, k: int) -> List[Tuple[Document, float]]:
        """
        Perform a similarity search with score, retrieving the top k documents.

        Args:
            query (str): The query to search for.
            k (int): The number of top documents to return.

        Returns:
            List[Tuple[Document, float]]: A list of Document and score tuples, sorted by relevance.
        """
        try:
            relevant_documents = self._get_relevant_documents(query)

            df = self.display_hits_as_df(relevant_documents)

            results = []
            for _, row in df.iterrows():
                doc = Document(page_content=row["body"], relevance=row["score"])
                results.append((doc, row["score"]))

            return sorted(results, key=lambda x: x[1], reverse=True)[:k]
        except Exception as e:
            print(f"Error in similarity search: {e}")
            return []



from indox.data_connectors.utils import Document
from typing import Any, List, Optional

class ArxivReader:
    """
    A reader class to access papers from the arXiv repository.
    """

    def __init__(self) -> None:
        """Initializes the ArxivReader object.

        Raises:
            ImportError: If the `arxiv` library is not installed.
        """
        try:
            import arxiv
        except ImportError:
            raise ImportError(
                "`arxiv` package not found, please run `pip install arxiv`"
            )

    @classmethod
    def class_name(cls) -> str:
        """Returns the class name ("ArxivReader")."""
        return "ArxivReader"

    def load_data(self, paper_ids: List[str], **load_kwargs: Any) -> List[Document] | Document:
        """Loads data from arXiv for the provided paper IDs.

        Args:
            paper_ids (List[str]): A list of arXiv paper IDs to read.

        Returns:
            List[Document]: A list of Document objects containing paper content and metadata.

        Raises:
            ValueError: If a paper ID is not found in the arXiv search results.
        """
        import arxiv

        documents = []
        for paper_id in paper_ids:
            try:
                search = arxiv.Search(id_list=[paper_id])
                paper = next(search.results())
            except StopIteration:
                raise ValueError(f"Paper ID '{paper_id}' not found in arXiv search results.")

            # Extract paper information
            title = paper.title
            abstract = paper.summary
            authors = ', '.join(author.name for author in paper.authors)

            # Combine information into content
            content = f"Title: {title}\n\nAuthors: {authors}\n\nAbstract: {abstract}"

            # Create metadata
            metadata = {
                "paper_id": paper_id,
                "title": title,
                "authors": authors,
                "published": paper.published,
                "arxiv_url": paper.entry_id,
            }

            # Create Document instance
            documents.append(Document(source="arXiv", content=content, metadata=metadata))
        if len(documents) == 1:
            return documents[0]

        return documents


    def load_content(self, paper_ids: List[str], **load_kwargs: Any) -> List[str] | str:
        """Loads paper content from arXiv for the provided paper IDs.

        Args:
            paper_ids (List[str]): A list of arXiv paper IDs to read.

        Returns:
            List[str]: A list of strings containing paper content (title, authors, abstract).

        Raises:
            ValueError: If a paper ID is not found in the arXiv search results.
        """
        import arxiv

        contents = []
        for paper_id in paper_ids:
            try:
                search = arxiv.Search(id_list=[paper_id])
                paper = next(search.results())
            except StopIteration:
                raise ValueError(f"Paper ID '{paper_id}' not found in arXiv search results.")

            # Extract paper information
            title = paper.title
            abstract = paper.summary
            authors = ', '.join(author.name for author in paper.authors)

            # Combine information into content
            content = f"Title: {title}\n\nAuthors: {authors}\n\nAbstract: {abstract}"

            # Append content to list
            contents.append(content)
        if len(contents) == 1:
            return contents[0]

        return contents



from typing import Optional, List, Tuple
from indox.core import Document
from indox.data_loader_splitter.SimpleLoadAndSplit.loader import create_document
from loguru import logger
import sys
from indox.splitter import semantic_text_splitter

# Set up logging
logger.remove()  # Remove the default logger
logger.add(sys.stdout,
           format="<green>{level}</green>: <level>{message}</level>",
           level="INFO")

logger.add(sys.stdout,
           format="<red>{level}</red>: <level>{message}</level>",
           level="ERROR")



def get_chunks(file_path, chunk_size, remove_sword):
    """
    Extract chunks from an unstructured document file using an unstructured data processing library.

    Parameters:
    - file_path (str): The path to the file containing unstructured data.
    - chunk_size (int, optional): The maximum size (in characters) for each chunk. Defaults to 500.

    Returns:
    - list: A list of `Document` objects, each containing a portion of the original content with relevant metadata.

    Raises:
    - Exception: Any errors that occur during document processing or chunking.

    Notes:
    - The function uses a title-based chunking method to segment the unstructured data into logical parts.
    - Metadata is cleaned and filtered to ensure proper structure before being added to the `Document` objects.
    - The `filter_complex_metadata` function is used to simplify and sanitize metadata attributes.

    """
    try:
        logger.info("Starting processing")

        # Create initial document elements using the 'PyPDF2' library
        text = create_document(file_path)
        logger.info("Created initial document elements")

        # split text by using 'bert-base-uncased' model and split it into chunks
        texts = semantic_text_splitter(text, max_tokens=chunk_size)
        for i in range(len(texts)):
            texts[i] = texts[i].replace("\n", " ")

        # Optionally remove stopwords from the chunks
        if remove_sword:
            from indox.data_loader_splitter.utils.clean import remove_stopwords_chunk
            texts = remove_stopwords_chunk(texts)


        logger.info("Completed chunking process")
        return texts

    except Exception as e:
        logger.error("Failed at step with error: %s", e)
        raise

class SimpleLoadAndSplit:
    def __init__(self, file_path: str, remove_sword: bool = False, max_chunk_size: int = 500):
        try:
            self.file_path = file_path
            self.remove_sword = remove_sword
            self.max_chunk_size = max_chunk_size
            logger.info("UnstructuredLoadAndSplit initialized successfully")
        except Exception as e:
            logger.error("Error initializing UnstructuredLoadAndSplit: %s", e)
            raise

    def load_and_chunk(self) -> List['Document']:
        """
        Split an unstructured document into chunks.

        Returns:
        - List[Document]: A list of `Document` objects, each containing a portion of the original content with relevant metadata.
        """
        try:
            docs = get_chunks(self.file_path, self.max_chunk_size, self.remove_sword)
            logger.info("Successfully obtained all documents")
            return docs
        except Exception as e:
            logger.error("Error in get_all_docs: %s", e)
            raise


from typing import List


def download():
    import nltk
    nltk.download('stopwords')
    nltk.download('punkt')


def remove_stopwords(text):
    from nltk.corpus import stopwords
    from nltk.tokenize import word_tokenize

    download()
    words = word_tokenize(text)
    stop_words = set(stopwords.words('english'))
    filtered_words = [word for word in words if word.lower() not in stop_words]
    filtered_text = ' '.join(filtered_words)
    return filtered_text


def remove_stopwords_chunk(chunks: List[str]) -> List[str]:
    return [remove_stopwords(a) for a in chunks]

#
# from typing import List
#
# nlp = None  # Initialize nlp as None
#
#
# def remove_stopwords(text: str) -> str:
#     """
#     Remove stopwords from a given text using spaCy.
#
#     Parameters:
#     - text (str): The input text from which stopwords will be removed.
#
#     Returns:
#     - str: The text after removing stopwords.
#     """
#     import spacy
#
#     global nlp
#     if nlp is None:
#         nlp = spacy.load('en_core_web_sm')  # Load the model only once
#
#     doc = nlp(text)
#     filtered_words = [token.text for token in doc if not token.is_stop]
#     filtered_text = ' '.join(filtered_words)
#     return filtered_text
#
#
# def remove_stopwords_chunk(chunks: List[str]) -> List[str]:
#     """
#     Remove stopwords from a list of text chunks using spaCy.
#
#     Parameters:
#     - chunks (List[str]): A list of text chunks.
#
#     Returns:
#     - List[str]: A list of text chunks with stopwords removed.
#     """
#     return [remove_stopwords(chunk) for chunk in chunks]



from indox.data_connectors.utils import Document
from typing import Any, List


class YoutubeTranscriptReader:
    """YouTube Transcript reader.

    This class retrieves transcripts from YouTube videos using the `youtube_transcript_api`
    package. It processes a list of YouTube video links and fetches the transcript text for each video.

    Attributes:
        languages (tuple): Tuple of language codes to fetch transcripts in. Defaults to English ("en").
    """

    languages: tuple = ("en",)

    @classmethod
    def class_name(cls) -> str:
        """Get the name of the class.

        Returns:
            str: The name of the class, "YoutubeTranscriptReader".
        """
        return "YoutubeTranscriptReader"

    def load_data(self, ytlinks: List[str], **load_kwargs: Any) -> List[Document] | Document:
        """Load transcripts from the provided YouTube video links.

        Args:
            ytlinks (List[str]): List of YouTube video links from which transcripts are to be fetched.
            **load_kwargs: Additional keyword arguments to pass to `YouTubeTranscriptApi.get_transcript`.

        Returns:
            List[Document]: A list of Document instances, each containing the transcript and metadata
            for a YouTube video.

        Raises:
            ImportError: If the `youtube_transcript_api` package is not installed.
            ValueError: If the YouTube link format is invalid and cannot extract the video ID.
            Exception: For other unexpected errors that may occur while fetching transcripts.
        """
        try:
            from youtube_transcript_api import YouTubeTranscriptApi
        except ImportError:
            raise ImportError(
                "`youtube_transcript_api` package not found. Please install it using `pip install youtube-transcript-api`."
            )

        documents = []
        for link in ytlinks:
            try:
                video_id = self._extract_video_id(link)
                srt = YouTubeTranscriptApi.get_transcript(video_id, languages=self.languages, **load_kwargs)
                transcript = "\n".join(chunk["text"] for chunk in srt)
                metadata = {
                    "video_id": video_id,
                    "link": link,
                    "language": self.languages,
                }
                documents.append(Document(source="YouTube", content=transcript, metadata=metadata))
            except ValueError as e:
                print(f"ValueError: {e}")
            except Exception as e:
                print(f"Unexpected error while processing link '{link}': {e}")
        if len(documents) == 1:
            return documents[0]

        return documents

    def load_content(self, ytlinks: List[str], **load_kwargs: Any) -> List[str] | str:
        """Load transcripts from the provided YouTube video links.

        Args:
            ytlinks (List[str]): List of YouTube video links from which transcripts are to be fetched.
            **load_kwargs: Additional keyword arguments to pass to `YouTubeTranscriptApi.get_transcript`.

        Returns:
            List[str]: A list of strings, each containing the transcript for a YouTube video.

        Raises:
            ImportError: If the `youtube_transcript_api` package is not installed.
            ValueError: If the YouTube link format is invalid and cannot extract the video ID.
            Exception: For other unexpected errors that may occur while fetching transcripts.
        """
        try:
            from youtube_transcript_api import YouTubeTranscriptApi
        except ImportError:
            raise ImportError(
                "`youtube_transcript_api` package not found. Please install it using `pip install youtube-transcript-api`."
            )

        transcripts = []
        for link in ytlinks:
            try:
                video_id = self._extract_video_id(link)
                srt = YouTubeTranscriptApi.get_transcript(video_id, languages=self.languages, **load_kwargs)
                transcript = "\n".join(chunk["text"] for chunk in srt)
                transcripts.append(transcript)
            except ValueError as e:
                print(f"ValueError: {e}")
            except Exception as e:
                print(f"Unexpected error while processing link '{link}': {e}")
        if len(transcripts) == 1:
            return transcripts[0]

        return transcripts

    def _extract_video_id(self, link: str) -> str:
        """Extract the video ID from a YouTube link.

        Args:
            link (str): The YouTube video link.

        Returns:
            str: The extracted video ID.

        Raises:
            ValueError: If the video ID cannot be extracted from the link.
        """
        try:
            video_id = link.split("?v=")[-1]
            if not video_id:
                raise ValueError("Invalid YouTube link format. Unable to extract video ID.")
            return video_id
        except IndexError:
            raise ValueError("Invalid YouTube link format. Unable to extract video ID.")



from typing import List, Any, Dict, Tuple
import numpy as np
from indox.core import Document


class Vearch:
    def __init__(self, embedding_function: Any, db_name: str, space_name: str = "text_embedding_space"):
        
        """
        Initialize Vearch with an external embedding function, a database name, and space definition.
        
        Args:
        embedding_function: Function or model that converts text to embeddings.
        db_name: Name of the database to store documents.
        space_name: Name of the space (default: "text_embedding_space").
        """
        from vearch.config import Config
        self.embedding_function = embedding_function
        self.db_name = db_name
        self.space_name = space_name
        self.data = []
        self.dim = len(self.embedding_function.embed_query("test"))
        self.space_schema = self.create_space_schema()

        self.config = Config(host="http://localhost:9001", token="secret")


    def create_space_schema(self) :
        from vearch.schema.field import Field
        from vearch.schema.space import SpaceSchema
        from vearch.utils import DataType, MetricType
        from vearch.schema.index import FlatIndex, ScalarIndex
        """Define space schema with changed field names."""
        text_field = Field(
            "docs",
            DataType.STRING,
            desc="docs field",
            index=ScalarIndex("docs_idx"),
        )
        embedding_field = Field(
            "embedding",
            DataType.VECTOR,
            FlatIndex("embedding_idx", MetricType.Inner_product),
            dimension=self.dim,
        )
        space_schema = SpaceSchema(
            "text_embedding_space",
            fields=[text_field, embedding_field],
        )
        return space_schema

    def create_space(self, space_name: str):
        """Create a new space."""
        space_schema = self.create_space_schema()
        return self.vc.create_space(self.db_name, space_schema)

    def create_database(self, db_name: str):
        """Create a new database."""
        ret = self.vc.create_database(db_name)
        return ret

    def add(self, docs: List[Document]):
        """
        Add a list of Document objects into the Vearch database.
        
        Args:
        docs: List of Document objects containing text and metadata.
        """
        for doc in docs:
            embedding = self.embedding_function.embed_query(doc)
            self.data.append({
                "text": doc,
                "embedding": embedding,
            })
        


    def similarity_search_with_score(self, query: str, k: int = 5) -> List[Tuple[Document, float]]:
        """
        Perform similarity search and return documents with scores.
        
        Args:
        query: The query text to search for.
        k: Number of top results to return (default: 5).
        
        Returns:
        List of tuples with Document and similarity score.
        """
        from sklearn.metrics.pairwise import cosine_similarity

        query_embedding = np.array(self.embedding_function.embed_query(query))

        results = []
        for doc in self.data:
            if 'embedding' in doc and 'text' in doc:
                doc_embedding = np.array(doc['embedding'])
                text = doc['text']

                similarity = cosine_similarity(doc_embedding.reshape(1, -1), query_embedding.reshape(1, -1))[0][0]

                document = Document(page_content=text, metadata=doc.get("metadata", {}))

                results.append((document, similarity))

        if not results:
            raise ValueError("No document embeddings found.")
        
        results = sorted(results, key=lambda x: x[1], reverse=True)

        return results[:k]



from typing import List, Optional


class RecursiveCharacterTextSplitter:
    """
    A class for splitting text into chunks recursively based on specified separators.

    This class implements a recursive algorithm to split text into chunks of a specified size,
    with an optional overlap between chunks. It uses a list of separators to determine where
    to split the text, starting with the first separator in the list and moving to the next
    if the current separator is not found in the text.

    Attributes:
        chunk_size (int): The maximum size of each chunk.
        chunk_overlap (int): The number of characters to overlap between chunks.
        separators (List[str]): A list of separators to use for splitting the text.
    """

    def __init__(
            self,
            chunk_size: int = 400,
            chunk_overlap: int = 50,
            separators: Optional[List[str]] = None
    ):
        """
        Initialize the RecursiveCharacterTextSplitter.

        Args:
            chunk_size (int): The maximum size of each chunk. Defaults to 4000.
            chunk_overlap (int): The number of characters to overlap between chunks. Defaults to 200.
            separators (Optional[List[str]]): A list of separators to use for splitting the text.
                                              Defaults to ["\n\n", "\n", ". ", " ", ""].
        """
        self.chunk_size = chunk_size
        self.chunk_overlap = chunk_overlap
        self.separators = separators or ["\n\n", "\n", ". ", " ", ""]

    def split_text(self, text: str) -> List[str]:
        """
        Split the input text into chunks.

        Args:
            text (str): The input text to be split.

        Returns:
            List[str]: A list of text chunks.
        """
        return self._split_text(text, self.separators)

    def _split_text(self, text: str, separators: List[str]) -> List[str]:
        """
        Recursively split the text using the given separators.

        Args:
            text (str): The text to be split.
            separators (List[str]): The list of separators to use for splitting.

        Returns:
            List[str]: A list of text chunks.
        """
        chunks = []
        separator = separators[-1]
        for sep in separators:
            if sep in text:
                separator = sep
                break

        splits = self._split_with_separator(text, separator)

        current_chunk = []
        current_length = 0

        for split in splits:
            split_length = len(split)

            if current_length + split_length <= self.chunk_size:
                current_chunk.append(split)
                current_length += split_length
            else:
                if current_chunk:
                    chunk = self._merge_splits(current_chunk, separator)
                    chunks.append(chunk)

                    if self.chunk_overlap > 0:
                        overlap_splits = self._get_overlap(current_chunk)
                        current_chunk = overlap_splits
                        current_length = sum(len(s) for s in overlap_splits)
                    else:
                        current_chunk = []
                        current_length = 0

                if split_length > self.chunk_size:
                    if len(separators) > 1:
                        nested_chunks = self._split_text(split, separators[1:])
                        chunks.extend(nested_chunks)
                    else:
                        chunks.append(split)
                    current_chunk = []
                    current_length = 0
                else:
                    current_chunk = [split]
                    current_length = split_length

        if current_chunk:
            chunk = self._merge_splits(current_chunk, separator)
            chunks.append(chunk)

        return chunks

    def _split_with_separator(self, text: str, separator: str) -> List[str]:
        """
        Split the text using the given separator.

        Args:
            text (str): The text to be split.
            separator (str): The separator to use for splitting.

        Returns:
            List[str]: A list of split text pieces, with separators included except for the last piece.
        """
        if separator:
            splits = text.split(separator)
            return [s + separator for s in splits[:-1]] + [splits[-1]]
        return list(text)

    def _merge_splits(self, splits: List[str], separator: str) -> str:
        """
        Merge the splits back into a single string.

        Args:
            splits (List[str]): The list of split text pieces to merge.
            separator (str): The separator to use when joining the splits.

        Returns:
            str: The merged text.
        """
        return separator.join(splits).strip()

    def _get_overlap(self, current_chunk: List[str]) -> List[str]:
        """
        Calculate the overlap for the next chunk.

        Args:
            current_chunk (List[str]): The current chunk of text splits.

        Returns:
            List[str]: A list of splits that form the overlap for the next chunk.
        """
        overlap_length = 0
        overlap_splits = []
        for s in reversed(current_chunk):
            if overlap_length + len(s) > self.chunk_overlap:
                break
            overlap_splits.insert(0, s)
            overlap_length += len(s)
        return overlap_splits



import uuid
import json
import numpy as np
from typing import List, Tuple, Callable
from indox.core import Document


class ApacheCassandra:
    """
    A class to interface with an Apache Cassandra-based vector store for storing and searching text data using embeddings.

    Attributes:
    - cluster_ips (List[str]): Fixed list of IP addresses of the Cassandra nodes.
    - port (int): Fixed port number to connect to the Cassandra cluster.
    - embedding_function (Callable[[str], np.ndarray]): A function that takes text input and returns an embedding vector.
    - keyspace (str): The keyspace to use in Cassandra.

    Methods:
    - __init__(embedding_function, keyspace):
        Initializes the ApacheCassandra instance with specified embedding function and keyspace.

    - add(docs):
        Adds a list of documents to the Cassandra vector store.

        Parameters:
        - docs (List[str]): A list of text documents to be added to the vector store.

        Raises:
        - ValueError: If `docs` is not a list of strings.
        - RuntimeError: If there is an issue with the Cassandra database operation.

    - similarity_search_with_score(query, k=5):
        Performs a similarity search on the Cassandra vector store with the provided query and returns the most similar documents along with their scores.

        Parameters:
        - query (str): The query text for which similar documents are to be found.
        - k (int, optional): The number of top results to return. Default is 5.

        Returns:
        - List[Tuple[Document, float]]: A list of tuples where each tuple contains a `Document` object and its corresponding similarity score.

        Raises:
        - RuntimeError: If there is an issue with the Cassandra database operation.

    - shutdown():
        Shuts down the Cassandra cluster connection.
    """

    def __init__(self, embedding_function: Callable[[str], np.ndarray], keyspace: str):
        from cassandra.cluster import Cluster
        from cassandra.policies import DCAwareRoundRobinPolicy
        try:
            self.cluster = Cluster(
                contact_points=['127.0.0.1'],
                port=9042,
                load_balancing_policy=DCAwareRoundRobinPolicy(local_dc='datacenter1'),
                protocol_version=4
            )
            self.session = self.cluster.connect()
            self.keyspace = keyspace
            self.embedding_function = embedding_function
            self._setup_keyspace()
        except Exception as e:
            raise RuntimeError(f"Failed to initialize Cassandra connection: {e}")

    def _setup_keyspace(self):
        try:
            self.session.execute(f"""
                CREATE KEYSPACE IF NOT EXISTS {self.keyspace}
                WITH REPLICATION = {{ 'class' : 'SimpleStrategy', 'replication_factor' : 1 }}
            """)
            self.session.set_keyspace(self.keyspace)
            self.session.execute("DROP TABLE IF EXISTS embeddings_table")
            self.session.execute(
                """CREATE TABLE IF NOT EXISTS embeddings_table (id UUID PRIMARY KEY, embedding text, chunk_text text)""")
        except Exception as e:
            raise RuntimeError(f"Failed to set up keyspace or table: {e}")

    def add(self, docs: List[str]):
        if not isinstance(docs, list):
            raise ValueError("The 'docs' argument must be a list of strings")
        try:
            chunk_embeddings = self.embedding_function.embed_documents(docs)
            for chunk, embedding in zip(docs, chunk_embeddings):
                embedding_str = json.dumps(embedding)
                self.session.execute("""INSERT INTO embeddings_table (id, embedding, chunk_text) VALUES (%s, %s, %s)""",
                                     (uuid.uuid4(), embedding_str, chunk))
        except Exception as e:
            raise RuntimeError(f"Failed to add documents to Cassandra: {e}")

    def similarity_search_with_score(self, query: str, k: int = 5) -> List[Tuple[Document, float]]:
        from sklearn.metrics.pairwise import cosine_similarity
        try:
            query_embedding = np.array(self.embedding_function.embed_query(query))
            rows = self.session.execute('SELECT id, embedding, chunk_text FROM embeddings_table')
            similarity_scores = []
            for row in rows:
                stored_embedding = np.array(json.loads(row.embedding))
                similarity = cosine_similarity(stored_embedding.reshape(1, -1), query_embedding.reshape(1, -1))[0][0]
                document = Document(page_content=row.chunk_text, id=str(row.id))
                similarity_scores.append((document, similarity))
            similarity_scores.sort(reverse=True, key=lambda x: x[1])
            return similarity_scores[:k]
        except Exception as e:
            raise RuntimeError(f"Failed to perform similarity search: {e}")

    def shutdown(self):
        try:
            self.cluster.shutdown()
        except Exception as e:
            raise RuntimeError(f"Failed to shutdown Cassandra cluster: {e}")



import warnings

# from .core import Document
# import logging
from .utils import show_indox_logo
from loguru import logger
import sys

warnings.filterwarnings("ignore")

# Set up logging
logger.remove()  # Remove the default logger
logger.add(sys.stdout,
           format="<green>{level}</green>: <level>{message}</level>",
           level="INFO")

logger.add(sys.stdout,
           format="<red>{level}</red>: <level>{message}</level>",
           level="ERROR")


class IndoxRetrievalAugmentation:
    def __init__(self):
        """
        Initialize the IndoxRetrievalAugmentation class
        """
        from . import __version__
        self.__version__ = __version__
        self.db = None

        self.multi_query_retrieval = None

        logger.info("IndoxRetrievalAugmentation initialized")
        show_indox_logo()

    class QuestionAnswer:
        def __init__(self, llm, vector_database, top_k: int = 5, document_relevancy_filter: bool = False,
                     generate_clustered_prompts: bool = False):
            self.qa_model = llm
            self.document_relevancy_filter = document_relevancy_filter
            self.top_k = top_k
            self.generate_clustered_prompts = generate_clustered_prompts
            self.vector_database = vector_database
            self.chat_history = {}
            self.context = []
            if self.vector_database is None:
                logger.error("Vector store database is not initialized.")
                raise RuntimeError("Vector store database is not initialized.")

        def invoke(self, query, multi_query: bool = False):
            if not query:
                logger.error("Query string cannot be empty.")
                raise ValueError("Query string cannot be empty.")

            try:
                if multi_query:
                    from indox.tools import MultiQueryRetrieval
                    self.multi_query_retrieval = MultiQueryRetrieval(self.qa_model, self.vector_database, self.top_k)

                    logger.info("Multi-query retrieval initialized")

                    return self.multi_query_retrieval.run(query)
                else:
                    logger.info("Retrieving context and scores from the vector database")
                    retrieved = self.vector_database._similarity_search_with_score(query, k=self.top_k)
                    context = [d[0].page_content for d in retrieved]
                    scores = [d[1] for d in retrieved]
                    if self.generate_clustered_prompts:
                        from .prompt_augmentation import generate_clustered_prompts
                        context = generate_clustered_prompts(context, embeddings=self.vector_database.embeddings,
                                                             summary_model=self.qa_model)

                    if not self.document_relevancy_filter:
                        logger.info("Generating answer without document relevancy filter")
                        answer = self.qa_model.answer_question(context=context, question=query)
                    else:
                        logger.info("Generating answer with document relevancy filter")
                        context = self.qa_model.grade_docs(context=context, question=query)
                        answer = self.qa_model.answer_question(context=context, query=query)

                    retrieve_context = context
                    key = len(self.chat_history)
                    new_entry = {'query': query, 'llm_response': answer, 'retrieval_context': context}
                    self.chat_history[key] = new_entry
                    self.context = retrieve_context
                    logger.info("Query answered successfully")
                    return answer
            except Exception as e:
                logger.error(f"Error while answering query: {e}")
                raise

    class AgenticRag:
        def __init__(self, llm, vector_database=None, top_k: int = 5):
            self.llm = llm
            self.top_k = top_k
            self.vector_database = vector_database
            self.chat_history = []
            self.context = []

        def run(self, query):
            grade_context = []
            if not query:
                logger.error("Query string cannot be empty.")
                raise ValueError("Query string cannot be empty.")
            else:
                if self.vector_database is not None:
                    context, scores = self.vector_database.retrieve(query, top_k=self.top_k)
                    grade_context = self.llm.grade_docs(context=context, question=query)
            # Any relevant doc? if yes -> next step(check for hallucination) , if no -> web search
            if len(grade_context) < 1:  # it means not relevant doc
                # go for web search
                logger.info("No Relevant document found, Start web search")
                from .utils import search_duckduckgo
                logger.info("No Relevant Context Found, Start Searching On Web...")
                results_for_searching_query = search_duckduckgo(query)
                logger.info("Answer Base On Web Search")
                answer = self.llm.answer_question(context=results_for_searching_query, question=query)
                logger.info("Check For Hallucination In Generated Answer Base On Web Search")

                hallucination_check_web_search_result = self.llm.check_hallucination(
                    context=results_for_searching_query, answer=answer)

                if hallucination_check_web_search_result.lower() == "yes":
                    logger.info("Hallucination detected, Regenerate the answer...")
                    # go for regenerate
                    answer = self.llm.answer_question(context=results_for_searching_query,
                                                      question=query)
                    return answer
                else:  # it means there is no hallucination
                    logger.info("Not Hallucinate")
                    return answer

            else:  # have relevant doc
                # check for hallucinating , if yes -> generate again , if no -> answer question
                answer = self.llm.answer_question(context=grade_context, question=query)
                hallucination_check = self.llm.check_hallucination(context=grade_context, answer=answer)
                if hallucination_check.lower() == "yes":
                    logger.info("Hallucination detected, Regenerate the answer...")
                    # go for regenerate
                    answer = self.llm.answer_question(context=grade_context,
                                                      question=query)
                    return answer
                else:  # it means there is no hallucination
                    logger.info("Not Hallucinate")
                    return answer




class MapsTextSearch:
    """
    Search for an address using OpenStreetMap's Nominatim API and handle printing.

    Parameters:
    - user_agent (str): The user agent string to be used by the Nominatim geolocator. Default is "osm_search".

    Methods:
    - search_address(address: str) -> None: Search for the given address and print the details.

    Notes:
    - The class uses the Geopy library to interact with the OpenStreetMap Nominatim API.
    - Results, including the full address, latitude, and longitude, are printed.
    """


    def __init__(self, user_agent: str = "osm_search"):
        """
        Initialize the MapsTextSearch with the specified user agent.

        Parameters:
        - user_agent (str): The user agent for the Nominatim geolocator. Default is "osm_search".
        """
        from geopy.geocoders import Nominatim
        from geopy.exc import GeopyError
        try:
            self.geolocator = Nominatim(user_agent=user_agent)
            print(f"MapsTextSearch initialized with user agent: {user_agent}")
        except GeopyError as e:
            print(f"Error initializing geolocator: {e}")
        except Exception as e:
            print(f"Unexpected error during initialization: {e}")

    def search_address(self, address: str) -> None:
        """
        Search for the given address and print the details.

        Parameters:
        - address (str): The address to search for.

        Returns:
        - None

        Notes:
        - Prints the full address, latitude, and longitude if the address is found.
        - Prints a message if the address is not found.
        - Prints an error message if an exception occurs during the search process.
        """
        from geopy.exc import GeopyError

        try:
            location = self.geolocator.geocode(address)
            if location:
                output = (f"Address: {location.address}\n"
                          f"Latitude: {location.latitude}, Longitude: {location.longitude}")
                print(output)
            else:
                print("Address not found")
        except GeopyError as e:
            print(f"Geopy error occurred: {e}")
        except Exception as e:
            print(f"Unexpected error occurred while searching for the address: {e}")
            self._log_fallback(f"Unexpected error: {e}")

    def _log_fallback(self, error_message: str) -> None:
        """
        Attempt to log an error to a fallback location if an unexpected error occurs.

        Parameters:
        - error_message (str): The error message to be logged.

        Returns:
        - None

        Notes:
        - Appends the error message to a fallback log file.
        """
        try:
            with open('fallback_log.txt', 'a') as fallback_log:
                fallback_log.write(f"{error_message}\n")
        except Exception as fallback_error:
            print(f"Error occurred while logging the original error: {fallback_error}")




# """
# This module provides the LLMGraphTransformer, GraphDocument, Node, and Relationship classes
# for building and managing graph structures from LLM outputs.
# """
#
# from .llmgraphtransformer import LLMGraphTransformer, GraphDocument, Node, Relationship
#
# __all__ = [
#     "LLMGraphTransformer",
#     "GraphDocument",
#     "Node",
#     "Relationship"
# ]
#



from typing import List
from indox.vector_stores.utils import filter_complex_metadata
from indox.core import Document


class PdfMiner:
    """
    Load a PDF file and extract its text and metadata using pdfminer.

    Parameters:
    - pdf_path (str): The path to the PDF file to be loaded.

    Methods:
    - load(): Reads the PDF file, extracts text, and creates a list of `Document` objects with metadata.

    Returns:
    - List[Document]: A list containing `Document` objects with the text content of each page and associated metadata.

    Raises:
    - FileNotFoundError: If the specified file does not exist.
    - RuntimeError: For any other errors encountered during PDF processing.
    """

    def __init__(self, pdf_path: str):
        from pdfminer.high_level import extract_text
        from pdfminer.pdfparser import PDFParser
        from pdfminer.pdfdocument import PDFDocument

        self.pdf_path = pdf_path
        self.pages: List[Document] = []
        self.metadata = {}
        self.extract_text = extract_text
        self.PDFParser = PDFParser
        self.PDFDocument = PDFDocument
        self.filter_complex_metadata = filter_complex_metadata

    def load(self) -> List[Document]:
        try:
            page_texts = self.extract_text(self.pdf_path).split('\f')

            with open(self.pdf_path, 'rb') as file:
                parser = self.PDFParser(file)
                document = self.PDFDocument(parser)
                self.metadata = document.info[0] if document.info else {}

            filtered_pdf_reader = self.filter_complex_metadata([self])[0]
            self.metadata = filtered_pdf_reader.metadata

            self.pages = [
                Document(page_content=page_text.strip(), **self.metadata)
                for page_text in page_texts
            ]

            return self.pages

        except FileNotFoundError:
            raise FileNotFoundError(f"The specified file '{self.pdf_path}' does not exist.")
        except Exception as e:
            raise RuntimeError(f"An error occurred while processing the PDF file: {e}")


    def load_and_split(self, splitter, remove_stopwords=False):
        from indox.data_loaders.utils import load_and_process_input
        return load_and_process_input(loader=self.load, splitter=splitter, remove_stopwords=remove_stopwords)




import uuid
from typing import Dict, Any, Optional

class Document:
    """Represents a data document from various sources.

    Attributes:
        id_: A unique identifier for the document.
        source: The source of the document, e.g., YouTube, Wikipedia.
        content: The actual content of the document.
        metadata: Additional metadata associated with the document.
    """

    id_: str = uuid.uuid4().hex  # Unique ID for each document
    source: str  # The source of the document, e.g., YouTube, Wikipedia
    content: str  # The actual content of the document
    metadata: Dict[str, Any]  # Metadata related to the document

    def __init__(self, source: str, content: str, metadata: Optional[Dict[str, Any]] = None):
        """Initializes a new Document object.

        Args:
            source: The source of the document.
            content: The content of the document.
            metadata: Optional metadata associated with the document.

        Raises:
            TypeError: If the source or content is not a string.
            ValueError: If the source or content is empty.
        """

        if not isinstance(source, str) or not source:
            raise ValueError("Source must be a non-empty string.")
        if not isinstance(content, str) or not content:
            raise ValueError("Content must be a non-empty string.")

        self.source = source
        self.content = content
        self.metadata = metadata or {}

    def __str__(self) -> str:
        """Returns a string representation of the document."""
        return f"Doc ID: {self.id_}\nSource: {self.source}\nContent: {self.content}\n"

    def to_dict(self) -> Dict[str, Any]:
        """Converts the document to a dictionary representation.

        Returns:
            A dictionary containing the document's attributes.
        """
        return {
            "doc_id": self.id_,
            "source": self.source,
            "content": self.content,
            "metadata": self.metadata,
        }

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "Document":
        """Creates a Document object from a dictionary.

        Args:
            data: A dictionary containing the document's attributes.

        Raises:
            KeyError: If the dictionary is missing required keys.
            TypeError: If the dictionary values are not of the expected types.
            ValueError: If the source or content is empty.

        Returns:
            A new Document object.
        """

        try:
            source = data["source"]
            content = data["content"]
            metadata = data.get("metadata", {})

            if not isinstance(source, str) or not source:
                raise ValueError("Source must be a non-empty string.")
            if not isinstance(content, str) or not content:
                raise ValueError("Content must be a non-empty string.")

            return cls(source=source, content=content, metadata=metadata)
        except KeyError as e:
            raise KeyError(f"Missing required key: {e.args[0]}")


from typing import List, Dict
from indox.core import Document

# Define the Neo4jGraph class for structured Cypher queries
class Neo4jGraph:
    def __init__(self, uri: str, username: str, password: str):
        """
        Initializes the connection to the Neo4j database.

        Args:
            uri (str): URI of the Neo4j instance (e.g., "bolt://localhost:7687").
            username (str): Username for Neo4j.
            password (str): Password for Neo4j.
        """
        try:
            # Import neo4j GraphDatabase dynamically to handle environments where it might not be installed.
            from neo4j import GraphDatabase
        except ImportError:
            raise ImportError(
                "Could not import the neo4j package. Please install it with `pip install neo4j`."
            )

        self.driver = GraphDatabase.driver(uri, auth=(username, password))

    def close(self):
        """Closes the Neo4j connection."""
        if self.driver:
            self.driver.close()

    def add_graph_documents(self, graph_documents: List['GraphDocument'], base_entity_label: bool = True, include_source: bool = True):
        """
        Adds graph documents to the Neo4j database.

        Args:
            graph_documents (List[GraphDocument]): List of GraphDocument objects to be added to the Neo4j database.
            base_entity_label (bool): Whether to add a base entity label to the node. Defaults to True.
            include_source (bool): Whether to include the source document as part of the graph. Defaults to True.
        """
        try:
            with self.driver.session() as session:
                for graph_doc in graph_documents:
                    self._add_graph_document(session, graph_doc, base_entity_label, include_source)
        finally:
            self.close()

    def _add_graph_document(self, session, graph_doc: 'GraphDocument', base_entity_label: bool, include_source: bool):
        """
        Adds a single GraphDocument to the Neo4j database.

        Args:
            session: Neo4j session.
            graph_doc (GraphDocument): The graph document to be added.
            base_entity_label (bool): Whether to add a base entity label to the node.
            include_source (bool): Whether to include the source document as part of the graph.
        """
        for node in graph_doc.nodes:
            self._add_node(session, node, base_entity_label)
        for relationship in graph_doc.relationships:
            self._add_relationship(session, relationship)
        if include_source:
            self._add_source(session, graph_doc)

    def _add_node(self, session, node: 'Node', base_entity_label: bool):
        """
        Adds a node to the Neo4j database.

        Args:
            session: Neo4j session.
            node (Node): The node object to be added.
            base_entity_label (bool): Whether to add a base entity label to the node.
        """
        labels = f":{node.type}"
        if base_entity_label:
            labels += ":Entity"
        query = f"MERGE (n{labels} {{id: $id}})"
        params = {"id": node.id}
        if node.embedding:
            query += " SET n.embedding = $embedding"
            params["embedding"] = node.embedding
        if node.text:
            query += " SET n.text = $text"
            params["text"] = node.text
        session.run(query, params)

    def _add_relationship(self, session, relationship: 'Relationship'):
        """
        Adds a relationship between two nodes in the Neo4j database.

        Args:
            session: Neo4j session.
            relationship (Relationship): The relationship to be added.
        """
        relationship_type = relationship.type.replace(" ", "_")
        query = f"""
        MATCH (a {{id: $source_id}}), (b {{id: $target_id}})
        MERGE (a)-[r:{relationship_type}]->(b)
        """
        session.run(query, {"source_id": relationship.source.id, "target_id": relationship.target.id})

    def _add_source(self, session, graph_doc: 'GraphDocument'):
        """
        Adds the source metadata to a special node in the Neo4j database.

        Args:
            session: Neo4j session.
            graph_doc (GraphDocument): The graph document containing the source information.
        """
        source = graph_doc.source
        query = """
        MERGE (s:Source {title: $title, page_content: $page_content})
        """
        params = {
            "title": source.metadata.get('metadata', {}).get('title', 'Untitled'),
            "page_content": source.page_content
        }
        session.run(query, params)

        for node in graph_doc.nodes:
            query = """
            MATCH (n {id: $node_id}), (s:Source {title: $title})
            MERGE (n)-[:HAS_SOURCE]->(s)
            """
            session.run(query, {"node_id": node.id, "title": source.metadata.get("metadata", {}).get("title", "Untitled")})

    def search_relationships_by_entity(self, entity_id: str, relationship_type: str):
        """
        Searches for relationships by entity ID and relationship type.

        Args:
            entity_id (str): The ID of the entity to search for.
            relationship_type (str): The type of relationship to search for.

        Returns:
            List[Dict]: A list of relationships including the source entity, relationship type, and target entity.
        """
        query = f"""
        MATCH (a {{id: $entity_id}})-[r:{relationship_type}]->(b)
        RETURN a, TYPE(r) AS rel_type, b
        """
        with self.driver.session() as session:
            result = session.run(query, {"entity_id": entity_id})
            return [record for record in result]


from indox.data_connectors.utils import Document
from typing import Any, List, Optional

class DiscordChannelReader:
    """Reads messages from a specified Discord channel using a Discord bot.

    **Important:** You need to create a Discord bot and get its token to use this class.
    See https://discord.com/developers/docs/intro for more information on creating a bot
    and obtaining its token.

    Attributes:
        bot_token: The bot token for your Discord bot.
        num_messages: The default number of messages to retrieve from a channel.
    """

    bot_token: str
    num_messages: Optional[int]

    def __init__(
        self,
        bot_token: str,
        num_messages: Optional[int] = 100,
    ) -> None:
        """Initializes the DiscordChannelReader object.

        Args:
            bot_token (str): The bot token for your Discord bot.
            num_messages (Optional[int]): The default number of messages to retrieve
                from a channel. Defaults to 100.

        Raises:
            ImportError: If the `discord.py` library is not installed.
        """
        try:
            import discord
        except ImportError:
            raise ImportError(
                "`discord` package not found, please run `pip install discord.py`"
            )

        self.bot_token = bot_token
        self.num_messages = num_messages

    @classmethod
    def class_name(cls) -> str:
        """Returns the class name ("DiscordChannelReader")."""
        return "DiscordChannelReader"

    def load_data(
        self,
        channel_ids: List[str],
        num_messages: Optional[int] = None,
        **load_kwargs: Any
    ) -> List[Document]:
        """Loads messages from specified Discord channels.

        Args:
            channel_ids (List[str]): A list of Discord channel IDs to read messages from.
            num_messages (Optional[int]): The maximum number of messages to retrieve
                from each channel. Defaults to the default value set during initialization
                or 100 if not set.

        Returns:
            List[Document]: A list of Document instances containing channel messages
                and metadata.

        Raises:
            ValueError: If a channel ID is not found in the Discord server.
        """
        import discord
        client = discord.Client(intents=discord.Intents.default())
        documents = []

        @client.event
        async def on_ready():
            for channel_id in channel_ids:
                channel = client.get_channel(int(channel_id))
                if channel:
                    messages = []
                    async for message in channel.history(limit=num_messages or self.num_messages):
                        messages.append(message.content)
                    response = "\n".join(messages)
                    metadata = {
                        "channel_id": channel_id,
                        "channel_name": channel.name,
                        "num_messages": len(messages),
                    }
                    documents.append(Document(source="Discord", content=response, metadata=metadata))
                else:
                    raise ValueError(f"Channel ID '{channel_id}' not found in Discord server.")

            await client.close()
        if len(documents) == 1:
            return documents[0]
        client.run(self.bot_token)
        return documents

    def load_content(
            self,
            channel_ids: List[str],
            num_messages: Optional[int] = None,
            **load_kwargs: Any
    ) -> List[str]:
        """Loads messages from specified Discord channels.

        Args:
            channel_ids (List[str]): A list of Discord channel IDs to read messages from.
            num_messages (Optional[int]): The maximum number of messages to retrieve
                from each channel. Defaults to the default value set during initialization
                or 100 if not set.

        Returns:
            List[str]: A list of strings, each containing the concatenated messages from a channel.

        Raises:
            ValueError: If a channel ID is not found in the Discord server.
        """
        import discord
        client = discord.Client(intents=discord.Intents.default())
        channel_contents = []

        @client.event
        async def on_ready():
            for channel_id in channel_ids:
                channel = client.get_channel(int(channel_id))
                if channel:
                    messages = []
                    async for message in channel.history(limit=num_messages or self.num_messages):
                        messages.append(message.content)
                    response = "\n".join(messages)
                    channel_contents.append(response)
                else:
                    raise ValueError(f"Channel ID '{channel_id}' not found in Discord server.")

            await client.close()
        client.run(self.bot_token)
        if len(channel_contents) == 1:
            return channel_contents[0]
        return channel_contents



import uuid
import json
from typing import Optional, List, Dict, Any, Tuple
from indox.core import Document

class LanternDB:
    DEFAULT_K = 4


    def __init__(
            self,
            collection_name: str,
            embedding_function: Optional[Any] = None,
            connection_params: Dict[str, str] = None,
            dimension: int = 768
    ) -> None:
        """
        Initialize the LanternDBClient class, connecting to LanternDB.

        Args:
            collection_name: Name of the collection to use in LanternDB.
            embedding_function: Function to generate embeddings.
            connection_params: Dictionary containing connection parameters for LanternDB.
            dimension: The dimensionality of the vector embeddings.
        """
        import psycopg2
        import psycopg2.extras

        self._collection_name = collection_name
        self._embedding_function = embedding_function
        self._dimension = dimension
        self._text_key = "text"
        self._vector_key = "embedding"

        self._conn = psycopg2.connect(**connection_params)

        print(f"Connected to LanternDB collection '{self._collection_name}'")
        self._cursor = self._conn.cursor(cursor_factory=psycopg2.extras.DictCursor)
        self._ensure_collection()

    def _ensure_collection(self):
        """
        Ensure the collection exists in LanternDB.
        """
        try:
            self._cursor.execute(f"""
                CREATE TABLE IF NOT EXISTS {self._collection_name} (
                    id TEXT PRIMARY KEY,
                    {self._text_key} TEXT,
                    {self._vector_key} VECTOR({self._dimension}),
                    metadata JSONB
                )
            """)
            self._conn.commit()
            print(f"Collection '{self._collection_name}' created successfully.")
        except Exception as e:
            self._conn.rollback()
            raise RuntimeError(f"Error ensuring collection: {e}")

    def add_texts(
            self,
            texts: List[str],
            metadatas: Optional[List[dict]] = None,
            ids: Optional[List[str]] = None
    ) -> List[str]:
        """
        Add texts and their embeddings to LanternDB.
        """
        if not ids:
            ids = [str(uuid.uuid4()) for _ in texts]

        if not self._embedding_function:
            raise ValueError("Embedding function must be provided")

        embeddings = self._embedding_function.embed_documents(texts)

        try:
            for i, text in enumerate(texts):
                metadata = metadatas[i] if metadatas else {}
                self._cursor.execute(f"""
                    INSERT INTO {self._collection_name} (id, {self._text_key}, {self._vector_key}, metadata)
                    VALUES (%s, %s, %s, %s)
                """, (ids[i], text, embeddings[i], json.dumps(metadata)))

            self._conn.commit()
            print(f"Inserted {len(texts)} documents into collection '{self._collection_name}'.")
        except Exception as e:
            self._conn.rollback()
            print(f"Error inserting documents: {e}")

        return ids

    def _similarity_search_with_score(
            self,
            query: str,
            k: int = DEFAULT_K,
            **kwargs: Any
    ) -> List[Tuple[Document, float]]:
        """
        Perform a similarity search in LanternDB and return documents with their scores.

        Args:
            query: The query string to search for.
            k: The number of similar texts to return.

        Returns:
            A list of tuples where each tuple contains a Document object and its similarity score.
        """
        if not self._embedding_function:
            raise ValueError("Embedding function is not set")

        query_embedding = self._embedding_function.embed_query(query)

        try:
            self._cursor.execute(f"""
                SELECT {self._text_key}, metadata, 1 - ({self._vector_key} <-> %s) AS similarity_score
                FROM {self._collection_name}
                ORDER BY similarity_score DESC
                LIMIT %s
            """, (query_embedding, k))

            results = self._cursor.fetchall()

            return [
                (
                    Document(
                        page_content=row[self._text_key],
                        metadata={
                            **json.loads(row['metadata']),
                            "Similarity Score": row['similarity_score'],
                        } if row['metadata'] else {"Similarity Score": row['similarity_score']},
                    ),
                    row['similarity_score']
                )
                for row in results
            ]
        except Exception as e:
            print(f"Error performing similarity search: {e}")
            return []

    def delete(self, ids: Optional[List[str]] = None):
        """Delete documents from the LanternDB collection."""
        try:
            if ids:
                self._cursor.execute(f"""
                    DELETE FROM {self._collection_name}
                    WHERE id = ANY(%s)
                """, (ids,))
                self._conn.commit()
                print(f"Deleted {len(ids)} documents from collection '{self._collection_name}'.")
            else:
                self._cursor.execute(f"DROP TABLE IF EXISTS {self._collection_name}")
                self._conn.commit()
                print(f"Deleted entire collection '{self._collection_name}'.")
        except Exception as e:
            self._conn.rollback()
            print(f"Error deleting documents: {e}")

    def __len__(self) -> int:
        """Return the count of documents in the LanternDB collection."""
        try:
            self._cursor.execute(f"SELECT COUNT(*) FROM {self._collection_name}")
            count = self._cursor.fetchone()[0]
            return count
        except Exception as e:
            print(f"Error getting document count: {e}")
            return 0

    def __del__(self):
        """Close the database connection when the object is destroyed."""
        if hasattr(self, '_cursor'):
            self._cursor.close()
        if hasattr(self, '_conn'):
            self._conn.close()



import os
from typing import Any, Dict, List, Optional, Union, Callable
from tenacity import retry_if_exception_type, stop_after_attempt, wait_exponential, before_sleep_log, retry
from indox.core import Embeddings
from loguru import logger
import sys

# Set up logging
logger.remove()  # Remove the default logger
logger.add(sys.stdout,
           format="<green>{level}</green>: <level>{message}</level>",
           level="INFO")

logger.add(sys.stdout,
           format="<red>{level}</red>: <level>{message}</level>",
           level="ERROR")

_ALLOW_REUSE_WARNING_MESSAGE = '`allow_reuse` is deprecated and will be ignored; it should no longer be necessary'


def _create_retry_decorator(max_retries: int) -> Callable[[Any], Any]:
    import cohere

    # support v4 and v5
    retry_conditions = (
        retry_if_exception_type(cohere.error.CohereError)
        if hasattr(cohere, "error")
        else retry_if_exception_type(Exception)
    )

    min_seconds = 4
    max_seconds = 10
    # Wait 2^x * 1 second between each retry starting with
    # 4 seconds, then up to 10 seconds, then 10 seconds afterwards
    return retry(
        reraise=True,
        stop=stop_after_attempt(max_retries),
        wait=wait_exponential(multiplier=1, min=min_seconds, max=max_seconds),
        retry=retry_conditions,
    )





class CohereEmbeddings(Embeddings):
    """Cohere embedding models.

    To use, you should have the `cohere` python package installed, and the
    environment variable `COHERE_API_KEY` set with your API key or pass it
    as a named parameter to the constructor.
    """

    client: Any  #: :meta private:
    async_client: Any  #: :meta private:
    model: str = "embed-english-v2.0"
    truncate: Optional[str] = None
    cohere_api_key: Optional[str] = None
    max_retries: int = 3
    request_timeout: Optional[float] = None
    user_agent: str = "langchain"

    class Config:
        extra = "forbid"

    def embed_with_retry(self, **kwargs: Any) -> Any:
        """Use tenacity to retry the embed call."""
        retry_decorator = _create_retry_decorator(self.max_retries)

        @retry_decorator
        def _embed_with_retry(**kwargs: Any) -> Any:
            return self.client.embed(**kwargs)

        return _embed_with_retry(**kwargs)

    def embed(self, texts: List[str], *, input_type: Optional[str] = None) -> List[List[float]]:
        """Embed a list of texts using the Cohere model.

        Args:
            texts: The list of texts to embed.
            input_type: The input type for the embedding (e.g., search_document, search_query).

        Returns:
            List[List[float]]: A list of embeddings for each text.
        """
        embeddings = self.embed_with_retry(
            model=self.model,
            texts=texts,
            input_type=input_type,
            truncate=self.truncate,
        ).embeddings
        return [list(map(float, e)) for e in embeddings]

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        """Embed a list of document texts.

        Args:
            texts: The list of texts to embed.

        Returns:
            List of embeddings, one for each text.
        """
        return self.embed(texts, input_type="search_document")

    def embed_query(self, text: str) -> List[float]:
        """Embed a single query text.

        Args:
            text: The text to embed.

        Returns:
            Embeddings for the text.
        """
        return self.embed([text], input_type="search_query")[0]



import warnings
from loguru import logger
import sys
from typing import List

warnings.filterwarnings("ignore")

logger.remove()
logger.add(sys.stdout,
           format="<green>{level}</green>: <level>{message}</level>",
           level="INFO")

logger.add(sys.stdout,
           format="<red>{level}</red>: <level>{message}</level>",
           level="ERROR")

class MultiQueryRetrieval:
    """
    A class that implements multi-query retrieval for enhanced information gathering.

    This class generates multiple queries from an original query, retrieves relevant
    information for each generated query, and combines the results to produce a final response.

    """

    def __init__(self, llm, vector_database, top_k: int = 3):
        """
        Initialize the MultiQueryRetrieval instance.

        Args:
            llm: The language model to use for query generation and response synthesis.
            vector_database: The vector database to use for information retrieval.
            top_k (int): The number of top results to retrieve for each query. Defaults to 3.
        """
        self.llm = llm
        self.vector_database = vector_database
        self.top_k = top_k

    def generate_queries(self, original_query: str) -> List[str]:
        """
        Generate multiple queries from the original query.

        Args:
            original_query (str): The original user query.

        Returns:
            List[str]: A list of generated queries.
        """
        prompt = f"Generate 3 different queries to gather information for answering the following question: {original_query}"
        response = self.llm.chat(prompt=prompt)
        return [q.strip() for q in response.split('\n') if q.strip()]

    def retrieve_information(self, queries: List[str]) -> List[str]:
        """
        Retrieve relevant information for each generated query.

        Args:
            queries (List[str]): A list of queries to use for information retrieval.

        Returns:
            List[str]: A list of relevant passages retrieved from the vector database.
        """
        all_relevants = []
        for query in queries:
            retrieved = self.vector_database._similarity_search_with_score(query, k=self.top_k)
            relevants = [d[0].page_content for d in retrieved]
            all_relevants.extend(relevants)
        return all_relevants

    def generate_response(self, original_query: str, context: List[str]) -> str:
        """
        Generate a final response based on the original query and retrieved context.

        Args:
            original_query (str): The original user query.
            context (List[str]): A list of relevant passages to use as context.

        Returns:
            str: The generated response.
        """
        combined_context = "\n".join(context)
        prompt = f"Based on the following information, answer the question: {original_query}\n\nContext: {combined_context}"
        return self.llm.chat(prompt=prompt)

    def run(self, query: str) -> str:
        """
        Execute the full multi-query retrieval process.

        This method orchestrates the entire process of query generation, information retrieval,
        and response generation.

        Args:
            query (str): The original user query.

        Returns:
            str: The final generated response.
        """
        logger.info(f"Running multi-query retrieval for: {query}")
        generated_queries = self.generate_queries(query)
        logger.info(f"Generated queries: {generated_queries}")

        relevants = self.retrieve_information(generated_queries)
        logger.info(f"Retrieved {len(relevants)} relevant passages")

        response = self.generate_response(query, relevants)
        logger.info("Generated final response")

        return response


import logging
from dspy import Signature, Module, Prediction,  OpenAI, settings

logging.basicConfig(level=logging.INFO,
                    format='%(asctime)s %(levelname)s:%(message)s')


class GenerateAnswer(Signature):
    """
    Signature class for generating an answer.

    Attributes:
        context (InputField): Contextual information helpful for answering the question.
        question (InputField): The question to be answered.
        answer (OutputField): The generated answer.
    """
    from dspy import  InputField, OutputField
    context = InputField(desc="Helpful information for answering the question.")
    question = InputField()
    answer = OutputField(desc="A detailed answer that is supported by the context. ONLY OUTPUT THE ANSWER!!")


class RAG(Module):
    """
    Module class for RAG (Retrieval-Augmented Generation) model.

    Attributes:
        generate_answer (ChainOfThought): Chain of thought for generating an answer.
        model (str): The model to be used.
        client (OpenAI): The OpenAI client.
    """
    def __init__(self, model, client):
        super().__init__()
        from dspy import ChainOfThought
        logging.info("Initializing RAG with model: %s", model)
        self.generate_answer = ChainOfThought(GenerateAnswer)
        self.model = model
        self.client = client

    def forward(self, context, question):
        """
        Forward method for the RAG model.

        Args:
            context (str): The context for the question.
            question (str): The question to be answered.

        Returns:
            Prediction: The prediction containing the context, question, and answer.
        """
        try:
            logging.info("Generating answer for question: %s", question)
            response = self.generate_answer(context=context, question=question, model=self.model, client=self.client)
            logging.info("Answer generated successfully")
            return Prediction(context=context, question=question, answer=response.answer)
        except Exception as e:
            logging.error("Error in RAG forward method: %s", e)
            raise


class DspyCot(Module):
    """
    Module class for DspyCotQA.

    Attributes:
        model (str): The model to be used.
        client (OpenAI): The OpenAI client.
    """

    def __init__(self, model, api_key):
        super().__init__()
        try:
            logging.info("Initializing DspyCotQA with model: %s", model)
            self.model = model
            self.client = OpenAI(model=self.model, api_key=api_key)
            settings.configure(lm=self.client)
            logging.info("DspyCotQA initialized successfully")
        except Exception as e:
            logging.error("Error initializing DspyCotQA: %s", e)
            raise

    def answer_question(self, context, question):
        """
        Answers a question given a context using the specified model.

        Args:
            context (str): The context for the question.
            question (str): The question to be answered.

        Returns:
            str: The answer to the question.
        """
        try:
            logging.info("Answering question: %s", question)
            rag = RAG(model=self.model, client=self.client)
            prediction = rag.forward(context, question)
            logging.info("Question answered successfully")
            return prediction.answer
        except Exception as e:
            logging.error("Error in DspyCotQA answer_question method: %s", e)
            return str(e)



from .mistral import Mistral
from .openai import OpenAi
# from .dspy import DspyCot
from .indox_api import IndoxApi
from .huggingface import HuggingFaceModel
from .googleai import GoogleAi
from .ollama import Ollama
# from .azure_openai import AzureOpenAi


from typing import Optional, Dict, Any, List, Tuple, Iterable, Union
from indox.core import VectorStore, Embeddings, Document
import uuid


# import pymongo
# from pymongo.collection import Collection


class MongoDB:
    DEFAULT_K = 4
    """
    A vector store implementation using MongoDB for document storage and retrieval.

    This class provides methods to add, retrieve, update, and delete documents in a MongoDB collection.
    It supports vector similarity search using cosine similarity when an embedding function is provided.

    Attributes:
        _client: MongoDB client instance.
        _db: MongoDB database instance.
        _collection: MongoDB collection instance.
        _embedding_function: Function to generate embeddings for documents.
        _index_name: Name of the index used for similarity search.
        _text_key: Key used to store document text in the collection.
        _embedding_key: Key used to store document embeddings in the collection.
    """

    def __init__(
            self,
            collection_name: str,
            embedding_function: Optional[Embeddings] = None,
            connection_string: str = "mongodb://localhost:27017/",
            database_name: str = "vector_db",
            index_name: str = "default",
            text_key: str = "text",
            embedding_key: str = "embedding",
    ) -> None:
        """
        Initialize the MongoDB vector store.

        Args:
            collection_name: Name of the MongoDB collection to use.
            embedding_function: Optional function to generate embeddings for documents.
            connection_string: MongoDB connection string.
            database_name: Name of the MongoDB database to use.
            index_name: Name of the index used for similarity search.
            text_key: Key used to store document text in the collection.
            embedding_key: Key used to store document embeddings in the collection.
        """
        from pymongo.errors import ConnectionFailure, ConfigurationError
        try:
            from pymongo import MongoClient
            self._client = MongoClient(connection_string)
            self._db = self._client[database_name]
            self._collection = self._db[collection_name]
        except ConnectionFailure:
            raise ConnectionFailure(f"Failed to connect to MongoDB server at {connection_string}")
        except ConfigurationError as e:
            raise ConfigurationError(f"MongoDB configuration error: {str(e)}")

        if not embedding_function:
            raise ValueError("Embedding function must be provided")
        self._embedding_function = embedding_function
        self._index_name = index_name
        self._text_key = text_key
        self._embedding_key = embedding_key

    @property
    def embeddings(self) -> Optional[Embeddings]:
        return self._embedding_function

    def _add_texts(
            self,
            texts: Iterable[str],
            metadatas: Optional[List[dict]] = None,
            ids: Optional[List[str]] = None,
            **kwargs: Any,
    ) -> List[str]:
        """
        Add texts to the vector store.

        Args:
            texts: Iterable of strings to add.
            metadatas: Optional list of metadata dictionaries.
            ids: Optional list of ids for the documents.
            **kwargs: Additional keyword arguments.

        Returns:
            List of ids for the added documents.
        """
        from pymongo.errors import BulkWriteError

        if ids is None:
            ids = [str(uuid.uuid4()) for _ in texts]

        if metadatas and len(metadatas) != len(texts):
            raise ValueError("Number of metadatas must match number of texts")

        embeddings = None
        if self._embedding_function:
            try:
                embeddings = self._embedding_function.embed_documents(list(texts))
            except Exception as e:
                raise ValueError(f"Error generating embeddings: {str(e)}")

        to_insert = []
        for i, text in enumerate(texts):
            doc = {
                "_id": ids[i],
                self._text_key: text,
            }
            if embeddings:
                doc[self._embedding_key] = embeddings[i]
            if metadatas and i < len(metadatas):
                doc["metadata"] = metadatas[i]
            to_insert.append(doc)

        try:
            self._collection.insert_many(to_insert)
        except BulkWriteError as e:
            raise BulkWriteError(f"Error inserting documents: {str(e)}")

        return ids

    def _add_documents(self, documents: List[Document], **kwargs: Any) -> List[str]:
        """
        Add documents to the vector store.

        Args:
            documents: List of Document objects to add.
            **kwargs: Additional keyword arguments.

        Returns:
            List of ids for the added documents.
        """
        texts = [doc.page_content for doc in documents]
        metadatas = [doc.metadata for doc in documents]
        return self._add_texts(texts, metadatas, **kwargs)

    def add(self, docs):
        """
        Adds documents to the MongoDB vector store.

        Args:
            docs: The documents to be added to the vector store.
        """
        try:
            if isinstance(docs[0], Document):
                self._add_documents(documents=docs)
            else:
                self._add_texts(texts=docs)
        except Exception as e:
            raise RuntimeError(f"Can't add document to the vector store: {e}")

    def _similarity_search_with_score(
            self,
            query: str,
            k: int = DEFAULT_K,
            filter: Optional[Dict[str, Any]] = None,
            **kwargs: Any,
    ) -> List[Tuple[Document, float]]:
        """
        Perform a similarity search and return documents with scores.

        Args:
            query: Query string.
            k: Number of results to return.
            filter: Optional filter to apply to the search.
            **kwargs: Additional keyword arguments.

        Returns:
            List of tuples containing Document objects and their similarity scores.
        """
        from sklearn.metrics.pairwise import cosine_similarity
        if not self._embedding_function:
            raise ValueError("Embedding function is not set")

        try:
            query_embedding = self._embedding_function.embed_query(query)
        except Exception as e:
            raise ValueError(f"Error generating query embedding: {str(e)}")

        try:
            if filter:
                docs = list(self._collection.find(filter, {self._embedding_key: 1, self._text_key: 1, "metadata": 1}))
            else:
                docs = list(self._collection.find({}, {self._embedding_key: 1, self._text_key: 1, "metadata": 1}))
        except Exception as e:
            raise RuntimeError(f"Error querying MongoDB: {str(e)}")

        if not docs:
            return []

        doc_embeddings = [doc[self._embedding_key] for doc in docs]
        similarities = cosine_similarity([query_embedding], doc_embeddings).flatten()

        scored_docs = list(zip(docs, similarities))

        scored_docs.sort(key=lambda x: x[1], reverse=True)
        top_k = scored_docs[:k]

        return [(Document(page_content=doc[self._text_key], metadata=doc.get("metadata", {})), score) for doc, score in
                top_k]
    def _similarity_search(
        self,
        query: str,
        k: int = 4,
        pre_filter: Optional[Dict] = None,
        post_filter_pipeline: Optional[List[Dict]] = None,
        **kwargs: Any,
    ) -> List[Document]:
        """Return MongoDB documents most similar to the given query.

        Uses the vectorSearch operator available in MongoDB Atlas Search.
        For more: https://www.mongodb.com/docs/atlas/atlas-vector-search/vector-search-stage/

        Args:
            query: Text to look up documents similar to.
            k: (Optional) number of documents to return. Defaults to 4.
            pre_filter: (Optional) dictionary of argument(s) to prefilter document
                fields on.
            post_filter_pipeline: (Optional) Pipeline of MongoDB aggregation stages
                following the vectorSearch stage.

        Returns:
            List of documents most similar to the query and their scores.
        """
        additional = kwargs.get("additional")
        docs_and_scores = self._similarity_search_with_score(
            query,
            k=k,
            pre_filter=pre_filter,
            post_filter_pipeline=post_filter_pipeline,
        )

        if additional and "similarity_score" in additional:
            for doc, score in docs_and_scores:
                doc.metadata["score"] = score
        return [doc for doc, _ in docs_and_scores]
    def delete(self, ids: Optional[List[str]] = None, **kwargs: Any) -> None:
        """
        Delete documents from the vector store.

        Args:
            ids: Optional list of document ids to delete.
            **kwargs: Additional keyword arguments.
        """
        if ids:
            self._collection.delete_many({"_id": {"$in": ids}})
        else:
            self._collection.delete_many({})

    def delete_collection(self) -> None:
        """Delete the entire collection from the database."""
        self._db.drop_collection(self._collection.name)

    def __len__(self) -> int:
        return self._collection.count_documents({})

    def get(
            self,
            ids: Optional[Union[str, List[str]]] = None,
            filter: Optional[Dict[str, Any]] = None,
            limit: Optional[int] = None,
            skip: Optional[int] = None,
            text_search: Optional[str] = None,
            include: Optional[List[str]] = None,
    ) -> Dict[str, Any]:
        """
        Retrieve documents from the vector store.

        Args:
            ids: Optional id or list of ids to retrieve.
            filter: Optional filter to apply.
            limit: Optional limit on the number of results.
            skip: Optional number of documents to skip.
            text_search: Optional text to search for.
            include: Optional list of fields to include in the results.

        Returns:
            Dictionary containing retrieved document information.
        """
        query = {}
        if ids:
            if isinstance(ids, str):
                query["_id"] = ids
            else:
                query["_id"] = {"$in": ids}

        if filter:
            query.update(filter)

        if text_search:
            query["$text"] = {"$search": text_search}

        projection = {"_id": 1}
        if include:
            if "embeddings" in include:
                projection[self._embedding_key] = 1
            if "metadata" in include:
                projection["metadata"] = 1
            if "text" in include:
                projection[self._text_key] = 1
        else:
            projection["metadata"] = 1
            projection[self._text_key] = 1

        cursor = self._collection.find(query, projection)

        if skip:
            cursor = cursor.skip(skip)
        if limit:
            cursor = cursor.limit(limit)

        results = list(cursor)
        return {
            "ids": [str(doc["_id"]) for doc in results],
            "embeddings": [doc.get(self._embedding_key) for doc in results] if "embeddings" in (
                    include or []) else None,
            "metadata": [doc.get("metadata") for doc in results],
            "documents": [doc.get(self._text_key) for doc in results],
        }

    def update_document(self, document_id: str, document: Document) -> None:
        """
        Update a single document in the vector store.

        Args:
            document_id: Id of the document to update.
            document: New Document object to replace the existing one.
        """
        self.update_documents([document_id], [document])

    def update_documents(self, ids: List[str], documents: List[Document]) -> None:
        """
        Update multiple documents in the vector store.

        Args:
            ids: List of document ids to update.
            documents: List of new Document objects to replace the existing ones.
        """
        if self._embedding_function is None:
            raise ValueError("For update, you must specify an embedding function on creation.")

        if len(ids) != len(documents):
            raise ValueError("Number of ids must match number of documents")

        text = [document.page_content for document in documents]
        metadata = [document.metadata for document in documents]
        try:
            embeddings = self._embedding_function.embed_documents(text)
        except Exception as e:
            raise ValueError(f"Error generating embeddings: {str(e)}")

        for i, doc_id in enumerate(ids):
            update_data = {
                self._text_key: text[i],
                self._embedding_key: embeddings[i],
                "metadata": metadata[i]
            }
            try:
                result = self._collection.update_one({"_id": doc_id}, {"$set": update_data})
                if result.modified_count == 0:
                    raise ValueError(f"No document found with id {doc_id}")
            except Exception as e:
                raise RuntimeError(f"Error updating document {doc_id}: {str(e)}")
